from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *

from unittest.case import TestCase

from sqlalchemy.orm.session import Session

from mocks import mock_db_session, mock
from zsl.application.containers.container import IoCContainer
from zsl.application.modules.alchemy_module import AlchemyModule
from zsl.application.modules.context_module import DefaultContextModule
from zsl.application.modules.logger_module import LoggerModule
from zsl.application.modules.task_router import TaskRouterModule
from zsl.service.service import SessionFactory
from zsl.testing.db import TestSessionFactory, DbTestModule, DbTestCase, IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration
from zsl import inject
from zsl.utils.injection_helper import bind


class TestContainerNoTestSessionFactory(IoCContainer):
    logger = LoggerModule
    database = AlchemyModule
    context = DefaultContextModule
    task_router = TaskRouterModule


class TestContainerTestSessionFactory(TestContainerNoTestSessionFactory):
    session_factory = DbTestModule


class DbTestCaseTest(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name="TestSessionFactoryTest",
        config_object=IN_MEMORY_DB_SETTINGS,
        container=TestContainerNoTestSessionFactory
    )

    class DbTest(DbTestCase, TestCase):
        def runTest(self):
            pass

    def test_db_test_case(self):
        test = DbTestCaseTest.DbTest()
        with mock.patch('zsl.testing.db.metadata') as mock_metadata:
            mock_sess = mock_db_session()
            test.setUp()
            test.tearDown()
            mock_metadata.create_all.assert_called_once()
            mock_sess.begin.assert_called_once()
            mock_sess.rollback.assert_called_once()
            mock_sess.close.assert_called_once()

    def test_db_test_case_session(self):
        test = DbTestCaseTest.DbTest()
        mock_tsf = mock.MagicMock()
        bind(TestSessionFactory, to=mock_tsf)
        test.setUp()
        mock_tsf.create_test_session.assert_called_once()
        mock_tsf.create_session.assert_not_called()


class TestSessionFactoryTest(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name="TestSessionFactoryTest",
        config_object=IN_MEMORY_DB_SETTINGS,
        container=TestContainerNoTestSessionFactory
    )

    def test_single_session(self):
        self.assertNotEqual(TestSessionFactory(), TestSessionFactory(),
                            "Two different instances may not be equal.")

        f = TestSessionFactory()
        f.create_test_session()
        self.assertEqual(f.create_session(), f.create_session(),
                         "Two results must be equal.")

        self.assertIsInstance(f.create_session(), Session, "Session must be correct.")


class DbTestModuleTest(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name="DbTestModuleTest",
        config_object=IN_MEMORY_DB_SETTINGS,
        container=TestContainerTestSessionFactory
    )

    def test_test_session_factory_is_injected(self):
        @inject(session_factory=SessionFactory)
        def get_session_factory(session_factory):
            return session_factory

        @inject(test_session_factory=TestSessionFactory)
        def get_test_session_factory(test_session_factory):
            return test_session_factory

        self.assertIsInstance(get_session_factory(), TestSessionFactory,
                              "Correct session factory must be returned")

        self.assertIsInstance(get_test_session_factory(), TestSessionFactory,
                              "Correct session factory must be returned")

        self.assertEqual(get_test_session_factory(), get_session_factory(),
                         "Test session factory and session factory must "
                         "be the same")



"""
:mod:`zsl.application.containers.web_container`
-----------------------------------------------
"""
from __future__ import unicode_literals

from zsl.application.containers.core_container import CoreContainer
from zsl.application.modules.web.web_context_module import WebContextModule


class WebContainer(CoreContainer):
    """Configuration for web application."""
    context = WebContextModule



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging
from typing import Any

import click

from injector import Binder, provides, singleton

from zsl import Zsl, inject
from zsl.application.initialization_context import InitializationContext
from zsl.application.initializers.web_initializer import WebInitializer
from zsl.application.modules.cli_module import ZslCli
from zsl.application.modules.context_module import DefaultContextModule, default_initializers
from zsl.utils.injection_helper import simple_bind

#: Initializers used in unit web applications
web_initializers = default_initializers + (WebInitializer,)


class WebCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> None

        @zsl_cli.cli.group(help='Web related tasks.')
        def web():
            pass

        @web.command(help="Run web server and serve the application")
        @click.option('--host', '-h', help="host to bind to", default='127.0.0.1')
        @click.option('--port', '-p', help="port to bind to", default=5000)
        @inject(web_handler=WebHandler)
        def run(web_handler, host, port):
            # type: (WebHandler, str, int)->None
            web_handler.run_web(host=host, port=port)

        self._web = web

    @property
    def web(self):
        return self._web


class WebHandler(object):
    @inject(flask=Zsl)
    def run_web(self, flask, host='127.0.0.1', port=5000, **options):
        # type: (Zsl, str, int, **Any)->None
        """Alias for Flask.run"""
        return flask.run(
            host=flask.config.get('FLASK_HOST', host),
            port=flask.config.get('FLASK_PORT', port),
            debug=flask.config.get('DEBUG', False),
            **options
        )


class WebContextModule(DefaultContextModule):
    """Adds web application context to current configuration."""

    def _create_context(self):
        logging.getLogger(__name__).debug("Creating web context.")
        return InitializationContext(initializers=web_initializers)

    @provides(interface=WebCli, scope=singleton)
    def provide_web_cli(self):
        return WebCli()

    @provides(interface=WebHandler, scope=singleton)
    def provide_web_handler(self):
        return WebHandler()

    def configure(self, binder):
        # type: (Binder) -> None
        super(WebContextModule, self).configure(binder)
        simple_bind(binder, WebCli, singleton)



"""
:mod:`zsl.application.containers.core_container`
------------------------------------------------
"""
from __future__ import unicode_literals

from zsl.application.containers.container import IoCContainer
from zsl.application.modules.cli_module import CliModule
from zsl.application.modules.logger_module import LoggerModule
from zsl.application.modules.alchemy_module import AlchemyModule
from zsl.application.modules.cache_module import RedisCacheInjectionModule
from zsl.application.modules.context_module import DefaultContextModule
from zsl.application.modules.task_router import TaskRouterModule


class CoreContainer(IoCContainer):
    """Modules for basic Zsl application."""
    logger = LoggerModule
    database = AlchemyModule
    cache = RedisCacheInjectionModule
    context = DefaultContextModule
    task_router = TaskRouterModule
    cli = CliModule



"""
:mod:`zsl.utils.background_task`
--------------------------------
"""
from __future__ import unicode_literals

from typing import Callable

from zsl import inject, Config
from zsl_client import GearmanService, RawTask, JsonTask
from functools import wraps
from zsl.utils.params_helper import required_params


__author__ = 'Peter Morihladko'


def background_task_method(task):
    """Decorate an object method as a background task (called with help of
    gearman).

    You have to create a task which will handle the gearman call. The
    method arguments will be encoded as JSON.

    :param task: name of the task
    :type task: str
    :return: decorated function
    """

    # TODO ako vysledok vrat nejaky JOB ID, aby sa dalo checkovat na pozadi
    # TODO vytvorit este vseobecny background_task nielen pre metody

    def decorator_fn(fn):

        gearman = None

        @inject(config=Config)
        def gearman_connect(config):
            # type: (Config) -> GearmanService
            if 'GEARMAN' not in config or 'host' not in config['GEARMAN'] or 'GEARMAN_TASK_NAME' not in config:
                raise Exception("Missing gearman settings (trying to use backgorund task)")

            gearman_host = (config['GEARMAN']['host'], config['GEARMAN']['port']) if config['GEARMAN']['port'] \
                else config['GEARMAN']['host']
            gearman_service = GearmanService({'HOST': [gearman_host], 'TASK_NAME': config['GEARMAN_TASK_NAME']})
            gearman_service.set_blocking(False)

            return gearman_service

        def get_gearman_client():
            # type: () -> GearmanService
            global gearman

            if not gearman:
                gearman = gearman_connect()

            return gearman

        @wraps(fn)
        def background_task_decorator(*args, **kwargs):
            # The first of the args is self.
            t = RawTask(task, dict(method=fn.__name__, args=args[1:], kwargs=kwargs))
            t_result = get_gearman_client().call(t, [JsonTask])
            return t_result.result

        background_task_decorator._background_fn = fn
        return background_task_decorator

    return decorator_fn


def get_background_method(obj, params):
    required_params(params, 'method', 'args', 'kwargs')

    method_name = params['method']
    args = params['args']
    kwargs = params['kwargs']

    method = getattr(obj, method_name)

    if not callable(method) or not hasattr(method, '_background_fn'):
        raise Exception('{0} is not a background task'.format(params['method']))

    method = getattr(method, '_background_fn')

    def original_fn():
        return method.__get__(obj, obj.__class__)(*args, **kwargs)

    return original_fn



import unittest
import types
from zsl.utils.command_dispatcher import CommandDispatcher

__author__ = 'peter'


class TestCommandDispatcher(unittest.TestCase):
    def setUp(self):
        self.dispatcher = CommandDispatcher()

    def test_add_function(self):

        def dummy_function():
            pass

        self.assertEqual(len(self.dispatcher.commands), 0, "Empty commands")

        self.dispatcher.add_function(dummy_function)

        self.assertEqual(len(self.dispatcher.commands), 1, "Command added")
        self.assertEqual(self.dispatcher.commands[dummy_function.__name__],
                         dummy_function,
                         "Command added")

    def test_command_decorator(self):

        def another_dummy_fn():
            pass

        decorated_fn = self.dispatcher.command(another_dummy_fn)

        self.assertEqual(decorated_fn,
                         another_dummy_fn,
                         "Decorator hasn't returned the same function")

        @self.dispatcher.command
        def dummy_function():
            pass

        self.assertEqual(self.dispatcher.commands[dummy_function.__name__],
                         dummy_function,
                         "Decorator didn't decorate")

    def test_command_execute(self):
        dispatcher = self.dispatcher

        @dispatcher.command
        def return_five():
            return 5

        @dispatcher.command
        def adder(a, b):
            return a + b

        @dispatcher.command
        def summer(n, **kwargs):
            return sum(map(lambda x: x*n, kwargs.values()))

        self.assertEqual(dispatcher.execute_command('return_five'),
                         5, "Command didn't execute properly")

        self.assertEqual(dispatcher.execute_command('adder', {'a': 1, 'b': 2}),
                         adder(1, 2),
                         "Command has not taken given arguments")

        # testparameters
        self.assertRaises(TypeError,
                          dispatcher.execute_command, 'adder', {'a': 1, 'b': 2, 'e': 3})
        self.assertRaises(TypeError, dispatcher.execute_command, 'adder')

        self.assertEquals(dispatcher.execute_command('summer', {'n': 2, '1': 1, '2': 2, '3': 3}),
                          sum([2, 4, 6]),
                          "Command doesn't support kwarg parameters")

        with self.assertRaises(KeyError):
            dispatcher.execute_command('error_command')

    def test_bound(self):
        dispatcher = self.dispatcher

        class C(object):
            def __init__(self, a, b):
                self._a = a
                self._b = b

            @dispatcher.command
            def get_sum(self, x, y):
                return self._a + self._b + x + y

        class D(object):
            def __init__(self, a, b):
                self._a = a
                self._b = b

            @dispatcher.command
            def get_other_sum(self, x, y):
                return self._a + self._b + x + y

        @dispatcher.command
        def dummy_fun():
            return 5

        c = C(1, 2)
        d = D(10, 20)

        bounded_dispatcher = dispatcher.bound(c)

        self.assertEqual(bounded_dispatcher.commands['get_sum'], c.get_sum,
                         "Bound has not bounded the method")
        self.assertEqual(bounded_dispatcher.execute_command('get_sum', {'x': 10, 'y': 200}),
                         c.get_sum(10, 200),
                         "Bound method has false result")

        self.assertIs(type(bounded_dispatcher.commands['get_other_sum']), types.FunctionType,
                      "Bound has bounded a foreign method")
        self.assertEqual(bounded_dispatcher.execute_command('get_other_sum', {'self': d, 'x': 1000, 'y': 300}),
                         d.get_other_sum(1000, 300),
                         "Bound has returned a false result for not bounded method")

        self.assertEqual(bounded_dispatcher.execute_command('dummy_fun'),
                         5, "Bound has returned a false result for normal function")



"""
:mod:`zsl.contrib.modules.alembic_module`
-----------------------------------------

Alembic module is responsible for handling database migrations for SqlAlchemy 
database backend. The complete documentation of the Alembic project, 
all the options and command may be found `Alembic website 
<http://alembic.zzzcomputing.com/en/latest/index.html>`_.

If you need database migrations just install `alembic` extra of `zsl` 
package. Then you may use `alembic` commands from ZSL cli interface via 
`alembic` command.

Just use a standard way of calling cli with a container having `AlembicModule`.

.. code-block:: python

    class MyApplicationContainer(WebContainer):
        alembic = AlembicModule


    @inject(zsl_cli=ZslCli)
    def main(zsl_cli: ZslCli) -> None:
        zsl_cli()

    
    if __name__ == "__main__":
        os.environ[SETTINGS_ENV_VAR_NAME] = ... or set_profile('my-profile')
        app = Zsl(__name__, modules=MyApplicationContainer.modules())
        main()

Configuration
~~~~~~~~~~~~~

To configure one needs to provide a valid `AlembicConfiguration` usually in 
`default_settings.py`

.. code-block:: python

    ALEMBIC = AlembicConfiguration(alembic_directory="...")
    
Then all the required files will be stored in the directory. Also the 
`alembic.ini` file will be stored there although this is not a standard way.

When using `AlembicCli` interface the alembic commands will be functional 
though.

A little setup is required so that the connection is opened correctly. In 
`alembic.ini` in the given directory remove the following line.

.. code-block:: ini

    sqlalchemy.url = driver://user:pass@localhost/dbname

Also set `script_location` to `.` so that it does not rely on absolute paths 

.. code-block:: ini

    script_location = .

Then in `env.py`, since we removed the `sqlalchemy.url` we need to use a ZSL 
engine created using the correct urls from settings. To do so define the 
routines `run_migrations_offline` and `run_migrations_online` so that they 
use a correct `Engine`. 

.. code-block:: python

    @inject(zsl_config=Config)
    def run_migrations_offline(zsl_config):
        url = zsl_config['DATABASE_URI']
        context.configure(url=url, target_metadata=target_metadata, literal_binds=True)
    
        with context.begin_transaction():
            context.run_migrations()


    @inject(engine=Engine)
    def run_migrations_online(engine):
        with engine.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=target_metadata
            )
    
            with context.begin_transaction():
                context.run_migrations()

 
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging
import os

import click
from click.core import Context
from flask.config import Config

from zsl.contrib.modules.alembic_config import AlembicConfiguration

try:
    from alembic.config import CommandLine
except ImportError:
    CommandLine = None
    logging.getLogger(__name__).exception(
        "Can not import alembic. Please install it first `pip install zsl ["
        "alembic]`.")
    raise
from injector import Module, singleton, Binder, provides

from zsl.application.modules.cli_module import ZslCli
from zsl import inject
from zsl.utils.injection_helper import simple_bind


class AlembicCli(object):
    """Alembic Cli interface support."""

    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> AlembicCli
        logging.getLogger(__name__).debug("Creating Alembic CLI.")

        @zsl_cli.cli.command(help='Run alembic maintenance tasks.',
                             context_settings=dict(
                                 ignore_unknown_options=True,
                                 allow_extra_args=True
                             ))
        @click.pass_context
        @inject(alembic_cfg=AlembicConfiguration)
        def alembic(ctx, alembic_cfg):
            # type: (Context, AlembicConfiguration) -> None
            is_initializing = len(ctx.args) and ctx.args[0] == 'init'
            alembic_directory = alembic_cfg.alembic_directory
            if is_initializing:
                cwd = None
                ctx.args.append(alembic_directory)
            else:
                cwd = os.getcwd()
                os.chdir(alembic_directory)
            CommandLine().main(ctx.args)
            if is_initializing:
                default_ini_path = 'alembic.ini'
                target_ini_path = os.path.join(alembic_directory, 'alembic.ini')
                os.rename(default_ini_path, target_ini_path)
            else:
                os.chdir(cwd)

        self._alembic = alembic

    @property
    def alembic(self):
        return self._alembic

    def __call__(self, *args, **kwargs):
        self._alembic()


class AlembicModule(Module):
    """Adds Alembic support for migrations."""

    ALEMBIC_CONFIG_NAME = 'ALEMBIC'

    @provides(AlembicConfiguration)
    @inject(config=Config)
    def provide_alembic_configuration(self, config):
        # type: (Config) -> AlembicConfiguration
        return config.get(AlembicModule.ALEMBIC_CONFIG_NAME)

    def configure(self, binder):
        # type: (Binder) -> None
        simple_bind(binder, AlembicCli, singleton)



"""
:mod:`zsl.cache.id_helper`
--------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from future.utils import with_metaclass
from builtins import object
import abc
from zsl.db.model.app_model_json_decoder import get_json_decoder
from zsl.db.model.app_model_json_encoder import AppModelJSONEncoder
import json
from zsl.db.model.app_model import AppModel, RELATED_FIELDS_HINTS,\
    RELATED_FIELDS_CLASS, RELATED_FIELDS
from zsl.utils.import_helper import fetch_class


def encoder_identity(x):
    return x


def decoder_identity(module_name, x):
    return x


def app_model_decoder_fn(model_key, x):
    class_name = model_key.split(":", 1)[0]
    hints = model_key_hint_extractor(model_key)
    return json.loads(x, cls=get_json_decoder(class_name, hints)) if x is not None else None


def app_model_encoder_fn(x):
    return json.dumps(x, cls=AppModelJSONEncoder)


def create_key_class_prefix(cls):
    return "{0}.{1}".format(cls.__module__, cls.__name__)


def create_key_object_prefix(obj):
    return create_key_class_prefix(obj.__class__)


def model_key_generator(model):
    # TODO: Test
    def get_app_model_related_fields(app_model, prefix=''):
        """
        Fetches the list of related fields of the model.
         - checks for the AppModel instances.
         - checks in the lists and tuples.
        """
        d = app_model.__dict__
        related = set()
        for key in d:
            v = d[key]
            if isinstance(v, AppModel):
                related.add(prefix + key + '=' + create_key_object_prefix(v))
                related.update(get_app_model_related_fields(v, prefix + key + '__'))
            elif isinstance(v, list) or isinstance(v, tuple):
                for val in v:
                    related.add(prefix + key + '=' + create_key_object_prefix(val))
                    related.update(get_app_model_related_fields(v[0], prefix + key + '__'))

        return related

    related = sorted(get_app_model_related_fields(model))
    if len(related):
        return "{0}:{1}:+{2}".format(create_key_object_prefix(model), model.get_id(), "+".join(related))
    else:
        return "{0}:{1}".format(create_key_object_prefix(model), model.get_id())

# TODO: Test


def model_key_hint_extractor(key):
    splitted_key = key.split(':')
    if len(splitted_key) == 2:
        return None
    # The first is empty
    related = splitted_key[2].split('+')[1:]
    hints = {RELATED_FIELDS: {}}
    for r in related:
        path = r.split('=')[0].split('__')
        cls = fetch_class(r.split('=')[1])
        h = hints
        for p in path[:-1]:
            h = h[RELATED_FIELDS][p][RELATED_FIELDS_HINTS]

        rh = h[RELATED_FIELDS]
        if path[-1] not in rh:
            rh[path[-1]] = {
                RELATED_FIELDS_CLASS: cls,
                RELATED_FIELDS_HINTS: {
                    RELATED_FIELDS: {
                    }
                }
            }

    return hints


class IdHelper(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    def gather_page(self, page_key, decoder=decoder_identity):
        pass

    @abc.abstractmethod
    def fill_page(self, page_key, data, timeout, encoder=encoder_identity, model_key_generator=model_key_generator):
        pass

    @abc.abstractmethod
    def check_page(self, page_key):
        pass

    @abc.abstractmethod
    def check_key(self, key):
        pass

    @abc.abstractmethod
    def get_key(self, key):
        pass

    @abc.abstractmethod
    def invalidate_key(self, key):
        pass

    @abc.abstractmethod
    def invalidate_keys_by_prefix(self, key_prefix):
        pass

    @abc.abstractmethod
    def set_key(self, key, value, timeout):
        pass



"""
:mod:`zsl.tasks.asl.version_task`
---------------------------------

Created on 24.12.2012

..moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

from builtins import object
import sqlalchemy

from zsl import inject, Zsl
from zsl.task.task_decorator import json_output


class VersionTask(object):
    """
    Shows the versions of ASL and the various used libraries.
    """

    @inject(app=Zsl)
    def __init__(self, app):
        self._app = app

    @json_output
    def perform(self, data):
        return {
            "ASL": self._app.VERSION,
            "SqlAlchemy": sqlalchemy.__version__
        }



"""
:mod:`zsl.application.modules.alchemy_module`
---------------------------------------------
"""
from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod
from builtins import object

import logging

from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker, Session
from sqlalchemy.engine.base import Engine

from injector import Module, provides, singleton

from zsl import Config, inject


class SessionHolder(object):
    def __init__(self, sess_cls):
        self._sess_cls = sess_cls

    def __call__(self):
        # type: () -> Session
        return self._sess_cls()


class SessionFactory(object):
    """Creates a db session with an open transaction."""

    @inject(session_holder=SessionHolder)
    def __init__(self, session_holder):
        # type: (SessionHolder)->None
        self._session_holder = session_holder

    def create_session(self):
        # type: ()->Session
        return self._session_holder()


class TransactionHolder(object):
    def __init__(self):
        self._orm = None
        self._transaction_callback = []
        self._in_transaction = False

    def has_session(self):
        # type: () -> bool
        return self._orm is not None

    @property
    def session(self):
        # type: () -> Session
        return self._orm

    @property
    def in_transaction(self):
        return self._in_transaction

    def inject_session(self, session):
        # type: (Session) -> None
        """Used to set the session in the @transactional decorator."""
        self._orm = session
        self._in_transaction = True

    def commit(self):
        logging.getLogger(__name__).debug("Commit.")
        self._orm.commit()

    def rollback(self):
        logging.getLogger(__name__).debug("Rollback.")
        self._orm.rollback()

    def close(self):
        logging.getLogger(__name__).debug("Close.")
        self._orm.close()
        self._orm = None
        self._in_transaction = False

    def append_transaction_callback(self, callback):
        self._transaction_callback.append(callback)

    def run_transaction_callbacks(self):
        callbacks = self._transaction_callback
        self._transaction_callback = []
        for c in callbacks:
            c()


class EmptyTransactionalHolder(object):
    def __init__(self):
        self._session = None

    @property
    def session(self):
        return self._session


class TransactionHolderFactory(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def create_transaction_holder(self):
        # type: ()->TransactionHolder
        pass


class DefaultTransactionHolderFactory(TransactionHolderFactory):
    def create_transaction_holder(self):
        # type: ()->TransactionHolder
        return TransactionHolder()


class AlchemyModule(Module):
    """Adds SQLAlchemy to current configuration."""

    @provides(Engine, scope=singleton)
    @inject(config=Config)
    def provide_engine(self, config):
        # type: (Config) -> Engine

        engine = create_engine(config['DATABASE_URI'],
                               **config['DATABASE_ENGINE_PROPS'])
        logging.debug("Created DB configuration to {0}.".
                      format(config['DATABASE_URI']))

        return engine

    @provides(SessionHolder, scope=singleton)
    @inject(engine=Engine)
    def provide_session_holder(self, engine):
        # type: (Engine) -> SessionHolder

        session = SessionHolder(sessionmaker(engine, autocommit=False,
                                             expire_on_commit=False))
        logging.debug("Created ORM session")

        return session

    @provides(TransactionHolderFactory, scope=singleton)
    def provide_transaction_holder_factory(self):
        return DefaultTransactionHolderFactory()



"""
:mod:`zsl.utils.file_helper`
----------------------------

.. moduleauthor:: Peter Morihladko <morihladko@atteq.com>, Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
import os
import errno


def makedirs(path):
    """Behaves like `mkdir -p <path>`. Without failure if the path exists."""
    
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



"""
:mod:`zsl.testing.http`
-----------------------
This module allows for querying the API via HTTP.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import *

import json
from typing import Any
from typing import Dict
from typing import Union

from flask.testing import FlaskClient
from typing import AnyStr

from flask.wrappers import Response

from zsl import inject, Zsl
from zsl.constants import HttpHeaders, MimeType
from zsl.utils.http import get_http_status_code_value


def json_loads(str_):
    # type: (AnyStr) -> Dict[str, str]
    """Parse JSON from Flask response which could be in bytes in Py3."""
    if isinstance(str_, bytes):
        str_ = str_.decode()

    return json.loads(str_)


class HTTPTestCase(object):
    """Extends TestCase with methods for easier testing of HTTP requests."""

    _DEFAULT_REQUEST_TASK_HEADERS = {
        HttpHeaders.CONTENT_TYPE.value: MimeType.APPLICATION_JSON.value
    }

    def requestTask(self, client, task, data, headers=None):
        # type: (FlaskClient, str, dict, dict)->Response
        """
        Request a task using POST and convert the given data to JSON.

        :param client: The client to ZSL which will be used for the request.
        :param task: Url which will be requested using POST method.
        :param data: Data which will be posted and first converted to JSON.
        :param headers: Dictionary of headers that'll be appended the
                        Content-Type: application/json header.
        :return: Flask response.
        """

        if headers is None:
            headers = {}

        headers.update(HTTPTestCase._DEFAULT_REQUEST_TASK_HEADERS)
        return client.post(task, data=json.dumps(data), headers=headers)

    def assertHTTPStatus(self, status, test_value, msg):
        # type: (Union[int, HTTPStatus], int, AnyStr) -> None
        """Assert HTTP status

        :param status: http status
        :param test_value: flask respond status
        :param msg: test message
        """
        status = get_http_status_code_value(status)
        self.assertEqual(status, test_value, msg)

    def assertJSONData(self, rv, data, msg):
        # type: (Response, Any, AnyStr) -> None
        data1 = self.extractResponseJSON(rv)
        self.assertEqual(data1, data, msg)

    def extractResponseJSON(self, rv):
        # type: (Response) -> Dict
        return json.loads(rv.data.decode())

    @inject(app=Zsl)
    def getHTTPClient(self, app):
        # type: (Zsl) -> FlaskClient
        return app.test_client()

    @inject(app=Zsl)
    def getRequestContext(self, app):
        return app.test_request_context()



"""
:mod:`zsl` -- zsl module
========================

Main service module.

   :platform: Unix, Windows
   :synopsis: The Atteq Service Layer. Service for exposing data to clients. Just provides DB access, feeds access and \
    other various aspects of service applications.

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

__version__ = '0.16.11'

from flask import Config

from injector import Module

from zsl.application.initialization_context import InitializationContext as ApplicationContext

from zsl.utils.injection_helper import inject
from zsl.application.service_application import ServiceApplication

Zsl = ServiceApplication

# placeholder for default value used in function declaration for arguments which will be injected
Injected = None



"""
:mod:`zsl.application.service_application`
------------------------------------------

The module contains the main Zsl class - :class:`.ServiceApplication`. It is
responsible for gluing everything together.

There is a simple profile support. Profile is usually tied to a configuration
file for an environment. Usually environments slightly differ in settings -
especially connection strings. These twists can be managed via additional
configuration files with the `cfg` extension usually placed in the `settings`
package. The name of the environment configuration file is controlled via
`ZSL_SETTINGS` environment variable.


.. data:: SETTINGS_ENV_VAR_NAME

    Name of the environment variable to be read for the profile configuration
    file.

"""
from __future__ import unicode_literals

import os
from typing import Callable

from flask import Flask, Config
from injector import Injector, Binder, singleton
from typing import Any

from zsl import __version__
from zsl._state import set_current_app
from zsl.utils.warnings import deprecated
from zsl.application.initialization_context import InitializationContext

#: Name of the environment variable to be read for the profile configuration.
SETTINGS_ENV_VAR_NAME = 'ZSL_SETTINGS'


def get_settings_from_profile(profile, profile_dir=None):
    # type: (str, Any)->str
    """"Returns the  configuration file path for the given profile.

    :param profile: Profile name to be used.
    :param profile_dir: The directory where the profile configuration file should reside. It
                        may be also a module, and then the directory of the module is used.
    :return: Configuration file path.
    """
    if profile_dir is None:
        import settings
        profile_dir = settings

    if hasattr(profile_dir, '__file__'):
        profile_dir = os.path.dirname(profile_dir.__file__)

    return os.path.join(profile_dir, '{0}.cfg'.format(profile))


def set_profile(profile):
    # type (str)->None
    """Sets the current profile.

    :param profile: The profile to be used."""
    os.environ[SETTINGS_ENV_VAR_NAME] = get_settings_from_profile(profile)


class ServiceApplication(Flask):
    """Atteq Service Flask application."""

    VERSION = __version__

    def __init__(self, import_name, static_path=None, static_url_path=None,
                 static_folder='static', template_folder='templates',
                 instance_path=None, instance_relative_config=False,
                 modules=None, config_object=None, version=None,
                 default_settings_module='settings.default_settings'):
        super(ServiceApplication, self).__init__(import_name, static_path,
                                                 static_url_path,
                                                 static_folder, template_folder,
                                                 instance_path,
                                                 instance_relative_config)
        self._dependencies_initialized = False
        self._default_settings_module = default_settings_module
        self._is_initialized = False
        self._injector = None
        self._worker = None
        self._configure(config_object)
        self._app_version = version
        if not modules:
            from zsl.application.containers.core_container import CoreContainer
            modules = CoreContainer.modules()
        self._configure_injector(modules)
        self._initialize()
        self._dependencies_initialized = True

    def __str__(self):
        return "ZSL(application={0}, zsl_version={1}, app_version={2})".format(
            self.name, self.VERSION,
            self._app_version)

    def _configure(self, config_object=None):
        # type: (Any) -> None
        """Read the configuration from config files.
        Loads the default settings and the profile settings if available.
        Check :func:`.set_profile`.

        :param config_object:
            This parameter is the configuration decscription may be a dict or
            string describing the module from which the configuration is used.
            Default is settings.default_settings.
        """

        if config_object:
            self.config.from_mapping(config_object)
        else:
            self.config.from_object(self._default_settings_module)

        zsl_settings = os.environ.get(SETTINGS_ENV_VAR_NAME)
        if zsl_settings is not None:
            self.config.from_envvar(SETTINGS_ENV_VAR_NAME)

    def _initialize(self):
        """Run the initializers."""
        ctx = self.injector.get(InitializationContext)
        ctx.initialize()

    def _register(self):
        """Register the current instance into application stack."""
        set_current_app(self)

    def _get_app_module(self):
        # type: () -> Callable
        """Returns a module which binds the current app and configuration.

        :return: configuration callback
        :rtype: Callable
        """

        def configure(binder):
            # type: (Binder) -> Callable
            binder.bind(ServiceApplication, to=self, scope=singleton)
            binder.bind(Config, to=self.config, scope=singleton)

        return configure

    def _configure_injector(self, modules):
        """Create the injector and install the modules.

        There is a necessary order of calls. First we have to bind `Config` and
        `Zsl`, then we need to register the app into the global stack and then
        we can install all other modules, which can use `Zsl` and `Config`
        injection.

        :param modules: list of injection modules
        :type modules: list
        """
        self._register()
        self._create_injector()
        self._bind_core()
        self._bind_modules(modules)
        self.logger.debug("Injector configuration with modules {0}.".format(modules))
        self._dependencies_initialized = True

    @deprecated
    def get_initialization_context(self):
        return self.injector.get(InitializationContext)

    def is_initialized(self):
        return self._dependencies_initialized

    @property
    def injector(self):
        # type: () -> Injector
        return self._injector

    @injector.setter
    def injector(self, value):
        self._injector = value

    def get_version(self):
        v = self.config.get('VERSION')
        if v is None:
            return ServiceApplication.VERSION
        else:
            return ServiceApplication.VERSION + ":" + v

    def _create_injector(self):
        self.logger.debug('Creating injector')
        self._injector = Injector()

    def _bind_core(self):
        self._injector.binder.bind(ServiceApplication, self, singleton)
        self._injector.binder.bind(Config, self.config, singleton)

    def _bind_modules(self, modules):
        for module in modules:
            self.injector.binder.install(module)



"""
:mod:`zsl.cache.redis_id_helper`
--------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

import logging

from zsl import inject, Config
from zsl.cache.id_helper import IdHelper, decoder_identity, encoder_identity, \
    model_key_generator
from zsl.cache.redis_cache_module import RedisCacheModule


class RedisIdHelper(IdHelper):
    CACHE_DEFAULT_TIMEOUT = 300

    @inject(config=Config, redis_cache_module=RedisCacheModule)
    def __init__(self, config, redis_cache_module):
        """
        Creates the id helper for caching support of AppModels.
        """
        self._config = config
        self._redis_cache_module = redis_cache_module

        if 'CACHE_TIMEOUTS' in self._config:
            self._cache_timeouts = self._config['CACHE_TIMEOUTS']
        else:
            self._cache_timeouts = None
        self._cache_default_timeout = self._config.get('CACHE_DEFAULT_TIMEOUT',
                                                       RedisIdHelper.CACHE_DEFAULT_TIMEOUT)

    def get_timeout(self, key, value, timeout):
        if self._cache_timeouts is None:
            return self._cache_default_timeout
        else:
            return self._cache_timeouts[timeout]

    def gather_page(self, page_key, decoder=decoder_identity):
        page_keys = self._redis_cache_module.get_list(page_key)
        logging.debug("Fetching page {0} from redis using keys {1}.".format(page_key, page_keys))

        p = []
        for k in page_keys:
            p.append(decoder(k, self.get_key(k)))

        return p

    def fill_page(self, page_key, data, timeout, encoder=encoder_identity, model_key_generator=model_key_generator):
        self._redis_cache_module.invalidate_key(page_key)
        first = True

        for d in data:
            key = model_key_generator(d)
            self._redis_cache_module.append_to_list(page_key, key)
            if first:
                self._redis_cache_module.set_key_expiration(page_key, self.get_timeout(page_key, data, timeout))
                first = False
            self._redis_cache_module.set_key(key, encoder(d), self.get_timeout(key, d, timeout))

    def check_page(self, page_key):
        if not self._redis_cache_module.contains_list(page_key):
            return False

        page_keys = self._redis_cache_module.get_list(page_key)
        for k in page_keys:
            if not self.check_key(k):
                return False

        return True

    def check_key(self, key):
        return self._redis_cache_module.contains_key(key)

    def get_key(self, key):
        return self._redis_cache_module.get_key(key)

    def invalidate_key(self, key):
        return self._redis_cache_module.invalidate_key(key)

    def invalidate_keys_by_prefix(self, key_prefix):
        return self._redis_cache_module.invalidate_by_glob(key_prefix + "*")

    def set_key(self, key, value, timeout):
        self._redis_cache_module.set_key(key, value, self.get_timeout(key, value, timeout))



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from enum import Enum


class MimeType(Enum):
    APPLICATION_JSON = 'application/json'
    # type: str


class HttpHeaders(Enum):
    CONTENT_TYPE = 'Content-Type'
    # type: str





"""
:mod:`zsl.application.containers.container`
-------------------------------------------
"""
from __future__ import unicode_literals

import inspect


class IoCContainer(object):
    """Collection of DI modules set as class attributes.

    This is used for declarative DI configuration, which can be easily
    extended upon.

    >>> class MyContainer(IoCContainer):
    ...   cache = MyCacheModule
    ...   redis = MyRedisModule

    >>> class MyTestContainer(MyContainer):
    ...   redis = MockRedisModule
    """
    @classmethod
    def modules(cls):
        """Collect all the public class attributes.

        All class attributes should be a DI modules, this method collects them
        and returns as a list.

        :return: list of DI modules
        :rtype: list[Union[Module, Callable]]
        """
        members = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a) and a.__name__ == 'modules'))
        modules = [module for name, module in members if not name.startswith('_')]
        return modules



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from collections import namedtuple

AlembicConfiguration = namedtuple('AlembicConfiguration', ['alembic_directory'])



"""
:mod:`zsl.db.helpers.query_filter`
----------------------------------
"""
from __future__ import unicode_literals
from builtins import object
from future.utils import viewitems
from sqlalchemy.sql import func


class OperatorEq(object):
    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr == v)


class OperatorNeq(object):
    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr != v)


class OperatorLike(object):
    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr.like('%{0}%'.format(v)))


class OperatorLeftLike(object):
    """
    Left side of string is like ...
    """

    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr.like('{0}%'.format(v)))


class OperatorRightLike(object):
    """
    Right side of string is like ...
    """

    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr.like('%{0}'.format(v)))


class OperatorBetween(object):
    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr.between(v[0], v[1]))


class OperatorCompareDates(object):
    """
    Compares only dates, year is not taken into account.
    Compared date value must be string in format '%m-%d'
    """

    @staticmethod
    def apply(q, attr, v):
        return q.filter(func.date_format(attr, '%m-%d') == v)


class RelationshipOperatorContains(object):
    @staticmethod
    def apply(q, attr, v):
        return q.filter(attr.contains(v))


FILTER_HINT = 'hint'
FILTER_VALUES = 'values'


class QueryFilter(object):
    """
    Helper class for applying filter criteria to query.
    """

    def __init__(self, query_filter, mappings=None, allow_null=False):
        """
        query_filter = {FILTER_VALUES: dict, FILTER_HINT: dict}
                - FILTER_VALUES dictionary (see example)
                - FILTER_HINTS dictionary - tells which operator (OperatorEq, OperatorBetween, ...)
                    use to which key from FILTER_VALUES dictionary
        mappings dict - maps keys from FILTER_VALUES to column attributes names of objects (see example)
                      - if the key from FILTER_VALUES is equal to the name of column attribute,
                        it doesn`t have to be mentioned in mappings
        allow_null boolean - if False (default value), None values from FILTER_VALUES will be ignored

        Example:
            query_filter = {
                FILTER_VALUES: {'fullname': 'jessica', 'lastname_initial': None},
                FILTER_HINT: {
                    'fullname': OperatorLike,
                    'lastname_initial': OperatorLeftLike,
                    'exclude_cid': OperatorNeq,
                }
            }
            mappings = {
                'lastname_initial': (Celebrity, 'lastname'),
                'exclude_cid': (Celebrity, 'cid'),
            }

            Notes: - 'fullname' key from FILTER_VALUES corresponds to column attribute Celebrity.fullname,
                    it doesn`t have to be mentioned in mappings
                  - if allow_null == False, key 'lastname_initial' from FILTER_VALUES will be ignored
                  - Celebrity is a sqlalechemy db model
        """
        if mappings is None:
            mappings = []

        self._query_filter = query_filter
        self._allow_null = allow_null
        self._mappings = mappings

    def apply_query_filter(self, q, cls):
        hints = self._query_filter[FILTER_HINT]
        values = self._query_filter[FILTER_VALUES]

        for k, v in viewitems(values):
            if v is None and not self._allow_null:
                continue

            if k in self._mappings:
                (cls, field) = self._mappings[k]
                attr = getattr(cls, field)
            else:
                attr = getattr(cls, k)

            q = hints[k]().apply(q, attr, v)

        return q



"""
:mod:`zsl.utils.params_helper`
------------------------------

Helpers for parameter handling

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals
from future.utils import viewitems

import inspect
from functools import reduce


class RequestException(Exception):
    """Exception raised on bad request"""
    pass


def required_params(data, *r_params):
    """Check if given parameters are in the given dict, if not raise an
    exception.

    :param data: data to check
    :type data: dict
    :param r_params: required parameters
    :raises RequestException: if params not in data
    """

    if not reduce(lambda still_valid, param: still_valid and param in data,
                  r_params, True):
        raise RequestException(msg_err_missing_params(*r_params))


def msg_err_missing_params(*params):
    return "Missing one or more required parameters (%s)" % '|'.join(params)


def safe_args(fn, args):
    """Check if ``args`` as a dictionary has the required parameters of ``fn``
    function and filter any waste parameters so ``fn`` can be safely called
    with them.

    :param fn: function object
    :type fn: Callable
    :param args: dictionary of parameters
    :type args: dict
    :return: dictionary to be used as named params for the ``fn``
    :rtype: dict
    """
    fn_args = inspect.getargspec(fn)

    if fn_args.defaults:
        required_params(args, fn_args.args[:-len(fn_args.defaults)])
    else:
        required_params(args, fn_args)

    if not fn_args.keywords:
        return {key: value for key, value in viewitems(args) if key in fn_args.args}
    else:
        return args



"""
:mod:`zsl.interface.task_queue`
-------------------------------

This module contains interfaces and functions to handle task queues in ZSL. A
task queue handles asynchronous and distributed code executions.

.. moduleauthor:: Peter Morihladko
"""
from future.utils import with_metaclass

import abc
import traceback
import socket

from typing import Any

from zsl import Zsl, Config, Injected
from zsl import inject
from zsl.router.task import TaskRouter
from zsl.task.job_context import JobContext, Job


class KillWorkerException(Exception):
    """If any task raises this exception a standalone worker will be killed."""
    pass


@inject(app=Zsl, task_router=TaskRouter)
def execute_job(job, app=Injected, task_router=Injected):
    # type: (Job, Zsl, TaskRouter) -> dict
    """Execute a job.

    :param job: job to execute
    :type job: Job
    :param app: service application instance, injected
    :type app: ServiceApplication
    :param task_router: task router instance, injected
    :type task_router: TaskRouter
    :return: task result
    :rtype: dict
    """

    app.logger.info("Job fetched, preparing the task '{0}'.".format(job.path))

    task, task_callable = task_router.route(job.path)
    jc = JobContext(job, task, task_callable)
    JobContext.set_current_context(jc)

    app.logger.info("Executing task.")
    result = jc.task_callable(jc.task_data)

    app.logger.info("Task {0} executed successfully.".format(job.path))

    return {'task_name': job.path, 'data': result}


class TaskQueueWorker(with_metaclass(abc.ABCMeta, object)):
    """Task queue worker abstraction.

    A task queue worker is responsible for communicating with a task queue and
    executing any task given by it.
    It should be able to run as a stand alone application.
    """

    @inject(app=Zsl, config=Config, task_router=TaskRouter)
    def __init__(self, app, config, task_router):
        # type: (Zsl, Config, TaskRouter) -> None
        self._app = app
        self._config = config
        self._task_router = task_router

        self._should_stop = False

        self._task_router.set_task_reloading(self._task_router.is_task_reloading() or self._config['RELOAD_GEARMAN'])

    @staticmethod
    def _get_client_id():
        # type: () -> str
        """Return client id.

        :return: client id
        :rtype: str
        """
        return "zsl-client-{0}".format(socket.gethostname())

    def handle_exception(self, e, task_path):
        # type: (Exception, str) -> dict
        """Handle exception raised during task execution.

        :param e: exception
        :type e: Exception
        :param task_path: task path
        :type task_path: str
        :return: exception as task result
        :rtype: dict
        """

        self._app.logger.error(str(e) + "\n" + traceback.format_exc())
        return {'task_name': task_path, 'data': None, 'error': str(e)}

    def execute_job(self, job):
        # type: (Job) -> dict
        """Execute job given by the task queue.

        :param job: job
        :type job: Job
        :return: task result
        :rtype: dict
        """
        try:
            return execute_job(job)

        except KillWorkerException:
            self._app.logger.info("Stopping Gearman worker on demand flag set.")
            self.stop_worker()

        except Exception as e:
            return self.handle_exception(e, job.path)

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        """Run the worker."""
        pass

    @abc.abstractmethod
    def stop_worker(self):
        """Stop the worker."""
        pass


@inject(worker=TaskQueueWorker)
def run_worker(self, worker, *args, **kwargs):
    # type: (TaskQueueWorker, *Any, **Any)->None
    """Run the app as a task queue worker.

    The worker instance is given as a DI module.
    """
    worker = self.injector.get(TaskQueueWorker)
    worker.run(*args, **kwargs)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *

try:
    import unittest.mock as mock
except ImportError:
    import mock

from zsl.service.service import SessionFactory
from zsl.testing.db import TestSessionFactory as DbTestTestSessionFactory
from zsl.utils.injection_helper import bind


def mock_db_session():
    mock_sess = mock.MagicMock()

    def session_holder():
        return mock_sess

    class TestSessionFactory(DbTestTestSessionFactory):
        def __init__(self):
            super(TestSessionFactory, self).__init__()
            self._session_holder = session_holder

    bind(SessionFactory, to=TestSessionFactory)
    bind(DbTestTestSessionFactory, to=TestSessionFactory)
    return mock_sess



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging

import click
from injector import Module, singleton, Binder

from zsl.interface.cli import cli
from zsl import inject
from zsl.utils.injection_helper import simple_bind
from zsl.utils.testing import load_and_run_tests


class ZslCli(object):
    def __init__(self):
        # type: () -> ZslCli
        logging.getLogger(__name__).debug("Creating ZSL CLI.")
        self._cli = cli

    @property
    def cli(self):
        # type: () -> ZslCli
        return self._cli

    def __call__(self, *args, **kwargs):
        # type: () -> None
        cli(**kwargs)


class ZslTaskCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> ZslTaskCli
        @zsl_cli.cli.command(help="Execute a single task.")
        @click.argument('task_path', metavar='task')
        @click.argument('data', default=None, required=False)
        def task(task_path, data=None):
            from zsl.interface.task import exec_task
            result = exec_task(task_path, data)
            click.echo(result)


class ZslTestCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> ZslTestCli
        @zsl_cli.cli.group(help="Perform unit tests.")
        def test():
            pass

        @test.command(help="run test")
        def run():
            load_and_run_tests()
            pass


class ZslGenerateCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> ZslGenerateCli
        @zsl_cli.cli.group(help="Perform unit tests.")
        def generate():
            pass

        @generate.command(help="Export python models to Backbone.js")
        @click.argument('module')
        @click.argument('models', metavar="model", nargs=-1)
        @click.option('--model-prefix', '-mp', help='namespace prefix for models (app.models.)')
        @click.option("--collection-prefix", "-cp", help="namespace prefix for collection (App.collections.)")
        @click.option("--model-fn", "-m", help="name of model constructor (MyApp.bb.Model)")
        @click.option("--collection-fn", "-c", help="name of collection constructor (MyApp.bb.Collection)")
        @click.option("--marker", help="marker to indicate the auto generated code")
        @click.option("--integrate", "-i", help="integrate to file", is_flag=True)
        @click.option('--js-file', '-j', help='path to file to integrate to')
        def javascript_models(module, models, model_prefix, collection_prefix,
                              model_fn, collection_fn, marker, integrate, js_file):
            from zsl.utils.deploy.js_model_generator import generate_js_models
            result = generate_js_models(module=module,
                                        models=models,
                                        model_prefix=model_prefix,
                                        collection_prefix=collection_prefix,
                                        model_fn=model_fn,
                                        collection_fn=collection_fn,
                                        marker=marker,
                                        integrate=integrate,
                                        js_file=js_file)
            click.echo(result)

        @generate.command(help="generate documentation for Apiary")
        def apiary_doc():
            from zsl.utils.deploy.apiari_doc_generator import generate_apiary_doc
            generator = generate_apiary_doc()
            click.echo(generator.get_doc())


class CliModule(Module):
    """Adds Alembic support for migrations."""

    def configure(self, binder):
        # type: (Binder) -> None
        bindings = [ZslCli, ZslTaskCli, ZslTestCli, ZslGenerateCli]
        for binding in bindings:
            simple_bind(binder, binding, singleton)



"""
Created on May 14, 2014

.. moduleauthor:: Peter Morihladko <morihladko@atteq.com>
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from unittest import TestCase

import pytest

from zsl import Zsl
from zsl.resource.model_resource import ModelResource
from resource.resource_test_helper import UserModel, \
    create_resource_test_data, users, UserTuple, UserAppModel
from zsl.testing.db import IN_MEMORY_DB_SETTINGS


@pytest.mark.skip(reason="Just a resource definition, no test.")
class TestRestResource(ModelResource):
    __model__ = UserModel


class ModelResourceTest(TestCase):
    def setUp(self):
        Zsl(__name__, config_object=IN_MEMORY_DB_SETTINGS)
        create_resource_test_data()

    def _testAppModel(self, user, model, msg=None):
        # type: (UserTuple, UserAppModel, str) -> None

        self.assertEqual(user.id, model.id, msg)
        self.assertEqual(user.name, model.name, msg)

    def testReadOne(self):
        resource = TestRestResource()
        test_user = users[0]
        user = resource.read(test_user.id)

        self._testAppModel(test_user, user, "resource value should be equal its model")

    def testReadListAll(self):
        resource = TestRestResource()
        res_users = resource.read(args={'limit': 'unlimited'})

        self.assertEqual(len(users), len(res_users), "resources count should equal model count")

        for i, d in enumerate(res_users):
            self._testAppModel(users[i], d, "list item value should equal model item value")

    def testCreate(self):
        resource = TestRestResource()
        user_data = {'name': 'eleven'}
        user_result = resource.create(params=None, args={}, data=user_data)
        user = resource.read(user_result.get_id())

        self.assertEqual(user_data['name'], user.name, "resource value should equal model value")

    def testUpdateOne(self):
        resource = TestRestResource()
        updated_name = 'updated'
        user = resource.read().pop()
        resource.update(params=user.get_id(), args={}, data={'name': updated_name})
        user = resource.read(user.get_id())

        self.assertEqual(updated_name, user.name, "updated resource value should equal model value")

    def testUpdateAll(self):
        resource = TestRestResource()
        updated_users = [UserTuple(id=1, name='updated_1'), UserTuple(id=2, name='updated_2')]
        resource.update(params=None, args={}, data=[x._asdict() for x in updated_users])

        for dt in updated_users:
            self._testAppModel(dt, resource.read(dt.id), "updated resource value should equal model value")

    def testDeleteOne(self):
        resource = TestRestResource()
        user = resource.read().pop()
        resource.delete(user.get_id(), args={}, data=None)
        res_users = resource.read(args={'limit': 'unlimited'})

        self.assertEqual(len(users) - 1, len(res_users), "count of resources should decline by 1")

    def testDeleteAll(self):
        resource = TestRestResource()
        resource.delete(params=None, args={}, data=None)
        res_users = resource.read()

        self.assertEqual(0, len(res_users), "count of resources should be 0")



"""
:mod:`zsl.router.task`
------------------------
"""
from __future__ import unicode_literals
from builtins import object

import importlib

from zsl import inject, Zsl, Config
from zsl.errors import ZslError
from zsl.utils.string_helper import underscore_to_camelcase
from zsl.utils.task_helper import instantiate, get_callable


class RoutingError(ZslError):
    pass


class TaskRouter(object):
    @inject(app=Zsl, config=Config)
    def __init__(self, app, config):
        # type: (Zsl, Config) -> None
        self._app = app
        self._config = config
        self._mappings = {}
        # Support for the settings with only one TASK_PACKAGE defined.
        self._task_packages = self._config.get('TASK_PACKAGES', (self._config.get('TASK_PACKAGE', 'zsl.tasks'),))
        self._debug = self._config['DEBUG']
        self._task_reloading = self._debug

    def get_task_packages(self):
        return self._task_packages

    def set_task_packages(self, task_packages):
        self._task_packages = task_packages

    def _load_module(self, module_name):
        # Debug loading provides better output.
        if self._debug:
            full = []
            for p in module_name.split('.'):
                full.append(p)
                importlib.import_module('.'.join(full))
        return importlib.import_module(module_name)

    def is_task_reloading(self):
        return self._task_reloading

    def set_task_reloading(self, value):
        self._task_reloading = value

    '''
    Returns the task handling the given request path.
    '''

    def route(self, path):
        self._app.logger.debug("Routing path '%s'.", path)

        # Split the path into arrays - package names in the tasks package.
        path = path.split("/")
        class_name = underscore_to_camelcase(path[-1])

        # finding the path in task packages
        module_ = None
        exceptions = []
        for task_package in self.get_task_packages():
            module_name = "{0}.{1}".format(task_package, ".".join(path))

            try:
                self._app.logger.debug("Trying to load module with name '%s' and class name '%s'.", module_name,
                                       class_name)
                module_ = self._load_module(module_name)
                break
            except ImportError as e:
                exceptions.append(e)
                if self._debug:
                    self._app.logger.warn(
                        "Could not load module with name '%s' and class name '%s', '%s'; proceeding to next module.",
                        module_name, class_name, e)

        if module_ is None:
            msg = "Can not load task named {0} from {1}.".format(".".join(path), ",".join(self.get_task_packages()))
            self._app.logger.warn(msg)
            for e in exceptions:
                self._app.logger.error("Reason", exc_info=e)
            raise RoutingError(msg)

        if self.is_task_reloading():
            importlib.reload(module_)

        # Create the task using the injector initialization.
        cls = getattr(module_, class_name)
        task = instantiate(cls)

        self._app.logger.debug("Task object {0} created [{1}].".format(class_name, task))
        return task, get_callable(task)



"""
:mod:`zsl.resource.json_server_resource`
----------------------------------------
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging
import re
from typing import Any

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import or_

from zsl.utils.http import get_http_status_code_value
from .model_resource import ModelResource
from zsl.interface.resource import ResourceResult
from zsl.service.service import transactional
from zsl.resource.resource_helper import filter_from_url_arg, flat_model, model_tree

import http.client
from flask import request


NOT_FOUND = ResourceResult(
    body={},
    status=get_http_status_code_value(http.client.NOT_FOUND)
)

NOT_IMPLEMENTED = ResourceResult(
    body={},
    status=get_http_status_code_value(http.client.NOT_IMPLEMENTED)
)

# any other arguments from these are considered as `property_name(_operator)=some_vaule` filter
_SKIPPED_ARGUMENTS = set(['callback', '_', 'q', '_start', '_end', '_sort', '_order', '_limit', '_embed', '_expand'])

# first group is the column name, then it can have a . separator or an operator suffix
_re_column_name = re.compile(r'^([^.]*?)(\..*?)?(_lte|_gte|_ne|_like)?$')


def _page_arg(p):
    # type: (int) -> str
    """Create a page argument from int."""
    return 'page=' + str(p)


def _get_link_pages(page, per_page, count, page_url):
    # type: (int, int, int, str) -> Dict[str, str]
    """Create link header for page metadata.

    :param page: current page
    :param per_page: page limit
    :param count: count of all resources
    :param page_url: url for resources
    :return: dictionary with name of the link as key and its url as value
    """
    current_page = _page_arg(page)
    links = {}
    end = page * per_page

    if page > 1:
        links['prev'] = page_url.replace(current_page, _page_arg(page - 1))

    if end < count:
        links['next'] = page_url.replace(current_page, _page_arg(page + 1))

    if per_page < count:
        links['first'] = page_url.replace(current_page, _page_arg(1))
        links['last'] = page_url.replace(current_page, _page_arg((count + per_page - 1) // per_page))

    return links


class JsonServerResource(ModelResource):
    """Model resource implementation to correspond with json-server.

    This implements the same REST interface which json-server
    (https://github.com/typicode/json-server) uses. It transforms the given
    input arguments into ModelResource-like and then adds metadata to result.
    """
    def to_filter(self, query, arg):
        """Json-server filter using the _or_ operator."""
        return filter_from_url_arg(self.model_cls, query, arg, query_operator=or_)

    def create(self, *args, **kwargs):
        """Adds created http status response and location link."""
        resource = super(JsonServerResource, self).create(*args, **kwargs)

        return ResourceResult(
            body=resource,
            status=get_http_status_code_value(http.client.CREATED),
            location="{}/{}".format(request.url, resource.get_id())
        )

    def _create_filter_by(self):
        """Transform the json-server filter arguments to model-resource ones."""
        filter_by = []

        for name, values in request.args.copy().lists():  # copy.lists works in py2 and py3
            if name not in _SKIPPED_ARGUMENTS:
                column = _re_column_name.search(name).group(1)

                if column not in self._model_columns:
                    continue

                for value in values:
                    if name.endswith('_ne'):
                        filter_by.append(name[:-3] + '!=' + value)
                    elif name.endswith('_lte'):
                        filter_by.append(name[:-4] + '<=' + value)
                    elif name.endswith('_gte'):
                        filter_by.append(name[:-4] + '>=' + value)
                    elif name.endswith('_like'):
                        filter_by.append(name[:-5] + '::like::%' + value + '%')
                    else:
                        filter_by.append(name.replace('__', '.') + '==' + value)

        filter_by += self._create_fulltext_query()

        return ','.join(filter_by)

    @staticmethod
    def _create_related(args):
        # type: (Dict) -> None
        """Create related field from `_embed` arguments."""
        if '_embed' in request.args:
            embeds = request.args.getlist('_embed')

            args['related'] = ','.join(embeds)

            del args['_embed']

    def _create_fulltext_query(self):
        """Support the json-server fulltext search with a broad LIKE filter."""
        filter_by = []

        if 'q' in request.args:
            columns = flat_model(model_tree(self.__class__.__name__, self.model_cls))
            for q in request.args.getlist('q'):
                filter_by += ['{col}::like::%{q}%'.format(col=col, q=q) for col in columns]

        return filter_by

    def _transform_list_args(self, args):
        # type: (dict) -> None
        """Transforms all list arguments from json-server to model-resource ones.

        This modifies the given arguments.
        """

        if '_limit' in args:
            args['limit'] = int(args['_limit'])
            del args['_limit']

        if '_page' in args:
            page = int(args['_page'])
            if page < 0:
                page = 1

            args['page'] = page
            del args['_page']

            if 'limit' not in args:
                args['limit'] = 10

        if '_end' in args:
            end = int(args['_end'])
            args['limit'] = end - int(args.get('_start', 0))

        if '_start' in args:
            args['offset'] = args['_start']
            del args['_start']

        if '_sort' in args:
            args['order_by'] = args['_sort'].replace('__', '.')
            del args['_sort']

            if args.get('_order', 'ASC') == 'DESC':
                args['order_by'] = '-' + args['order_by']

        if '_order' in args:
            del args['_order']

        filter_by = self._create_filter_by()
        if filter_by:
            args['filter_by'] = filter_by

    def read(self, params, args, data):
        """Modifies the parameters and adds metadata for read results."""
        result_count = None
        result_links = None

        if params is None:
            params = []
        if args:
            args = args.copy()
        else:
            args = {}

        ctx = self._create_context(params, args, data)
        row_id = ctx.get_row_id()

        if not row_id:
            self._transform_list_args(args)

            if 'page' in args or 'limit' in args:
                ctx = self._create_context(params, args, data)

                result_count = self._get_collection_count(ctx)

            if 'page' in args:
                result_links = _get_link_pages(
                    page=args['page'],
                    per_page=int(args['limit']),
                    count=result_count,
                    page_url=request.url
                )

            if 'limit' not in args:
                args['limit'] = 'unlimited'

        self._create_related(args)

        try:
            return ResourceResult(
                body=super(JsonServerResource, self).read(params, args, data),
                count=result_count,
                links=result_links
            )
        except NoResultFound:
            return NOT_FOUND

    def update(self, *args, **kwargs):
        """Modifies the parameters and adds metadata for update results.

        Currently it does not support `PUT` method, which works as replacing
        the resource. This is somehow questionable in relation DB.
        """
        if request.method == 'PUT':
            logging.warning("Called not implemented resource method PUT")

        resource = super(JsonServerResource, self).update(*args, **kwargs)

        if resource:
            return resource
        else:
            return NOT_FOUND

    @transactional
    def delete(self, params, args, data):
        """Supports only singular delete and adds proper http status."""
        ctx = self._create_context(params, args, data)
        row_id = ctx.get_row_id()

        if row_id:
            deleted = self._delete_one(row_id, ctx)

            if deleted:
                return ResourceResult(body={})
            else:
                return NOT_FOUND
        else:
            return NOT_FOUND




"""
:mod:`zsl.__main__` -- zsl main module
======================================

Main service module.
"""
from __future__ import unicode_literals
from zsl.interface.cli import cli

if __name__ == "__main__":
    cli()



"""
:mod:`zsl.tasks.asl.sum_task`
-----------------------------

Created on 22.12.2012

..moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

from builtins import object
from zsl.task.task_decorator import json_input, json_output
from zsl import Zsl
from injector import inject


class SumTask(object):
    @inject(app=Zsl)
    def __init__(self, app):
        self._app = app

    @json_input
    @json_output
    def perform(self, data):
        self._app.logger.debug("Sum task with data '{0}'.".format(data.get_data()))
        return {"input": data.get_data(), "result": sum(data.get_data())}



"""
:mod:`zsl.utils.type_helper`
----------------------------

Helper module for handling types.

.. moduleauthor:: peter
"""
from __future__ import unicode_literals


def not_empty_list(l):
    """Check if given value is a non-empty list.

    :param l: value
    :return: True on non-empty list, False otherwise
    :rtype: bool

    :Example:
        >>> not_empty_list([1, 2, 3])
        True
        >>> not_empty_list(None)
        False
    """
    return isinstance(l, list) and len(l) > 0



"""
:mod:`zsl.application.containers.celery_container`
--------------------------------------------------
"""
from __future__ import unicode_literals

from zsl.application.containers.core_container import CoreContainer
from zsl.application.modules.celery_module import CeleryTaskQueueOutsideWorkerModule, CeleryTaskQueueMainWorkerModule, \
    CeleryCliModule


class CeleryContainer(CoreContainer):
    """Configuration for celery application."""
    worker = CeleryTaskQueueMainWorkerModule
    celery_cli = CeleryCliModule


class CeleryStandAloneContainer(CoreContainer):
    """Configuration for application run with help of celery cli tools."""
    worker = CeleryTaskQueueOutsideWorkerModule
    celery_cli = CeleryCliModule



"""
:mod:`zsl.interface.gearman.task_filler`
----------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import print_function
from __future__ import unicode_literals


import gearman
import json

from zsl import Config, inject, Injected


@inject(config=Config)
def exec_task_filler(task_path, json_data, config=Injected):
    print("Initializing client.")
    gm_client = gearman.GearmanClient(["{0}:{1}".format(config['GEARMAN']['host'], config['GEARMAN']['port'])])
    print("Client initialized.")

    if json_data.startswith("file://"):
        with open(json_data[len("file://"):]) as f:
            json_data = f.read()

    ret_val = gm_client.submit_job(config['GEARMAN_TASK_NAME'], json.dumps(
        {'path': task_path, 'data': json.loads(json_data)}))

    print("Returned value '{0}'.".format(ret_val))
    output = json.loads(ret_val.result)['data']
    print("Returned data '{0}' of type '{1}'.".format(output, type(output)))




"""
:mod:`zsl.interface.celery.worker`
----------------------------------

Implementation of celery workers.

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals

import sys

from celery import Celery, shared_task
from injector import Module, singleton, provides

from zsl import inject
from zsl.interface.task_queue import TaskQueueWorker
from zsl.task.job_context import Job


class CeleryTaskQueueWorkerBase(TaskQueueWorker):
    """Base class for celery task queue worker.

    It contains only the task execution part of worker.
    """

    def execute_celery_task(self, job_data):
        # type: (dict) -> dict
        """Creates job from task and executes the job.

        :param job_data: job's data
        :return: job results
        :rtype: dict
        """
        job = Job(job_data)

        return self.execute_job(job)


class CeleryTaskQueueOutsideWorker(CeleryTaskQueueWorkerBase):
    """Celery worker used only for task execution.

    This can be used when the worker is manage with some other application,
    like `celery worker` or `celery multi`.
    """

    def stop_worker(self):
        self._app.logger.error("Running from celery worker, kill from shell!")

    def run(self):
        self._app.logger.error("Running from celery worker, start from shell!")


class CeleryTaskQueueMainWorker(CeleryTaskQueueWorkerBase):
    """Worker implementation for Celery task queue."""

    def __init__(self,):
        super(CeleryTaskQueueMainWorker, self).__init__()

        self.celery_app = Celery()
        self.celery_app.config_from_object(self._config['CELERY'])
        self.celery_worker = None

    def stop_worker(self):
        self._app.logger.info("Stopping Celery worker on demand - quitting.")
        self.celery_worker.stop()

    def run(self, argv, *args, **kwargs):
        self._app.logger.info("Running the worker.")
        self.celery_worker = self.celery_app.worker_main((sys.argv[0],) + argv)


@shared_task
@inject(worker=CeleryTaskQueueWorkerBase)
def zsl_task(job_data, worker):
    # type: (dict, CeleryTaskQueueWorkerBase) -> dict
    """This function will be registered as a celery task.

    :param job_data: task data
    :param worker: current celery worker
    :type worker: CeleryTaskQueueWorkerBase
    :return: task results
    """
    return worker.execute_celery_task(job_data)




"""
:mod:`zsl.service.service`
--------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>,
                  Peter Morihladko <morihladko@atteq.com>
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *  # NOQA

import logging
from functools import wraps

from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.session import Session
from zsl.application.modules.alchemy_module import TransactionHolderFactory, EmptyTransactionalHolder, SessionFactory

from zsl import inject, Zsl

_EMPTY_TX_HOLDER = EmptyTransactionalHolder()


class TransactionalSupportMixin(object):
    """This mixin allows the objects to access transactional holder."""

    @property
    def _orm(self):
        # type: ()->Session
        return getattr(self, _TX_HOLDER_ATTRIBUTE, _EMPTY_TX_HOLDER).session


class Service(TransactionalSupportMixin):
    """
    Main service class.
    """

    @inject(app=Zsl, engine=Engine)
    def __init__(self, app, engine):
        """Constructor - initializes and injects the needed libraries."""
        super(Service, self).__init__()
        self._app = app
        self._engine = engine


_TX_HOLDER_ATTRIBUTE = '_tx_holder'


def transactional(f):
    @inject(session_factory=SessionFactory)
    def _get_session_factory(session_factory):
        # type: (SessionFactory) -> SessionFactory
        return session_factory

    @inject(tx_holder_factory=TransactionHolderFactory)
    def _get_tx_holder_factory(tx_holder_factory):
        # type: (TransactionHolderFactory) -> TransactionHolderFactory
        return tx_holder_factory

    @wraps(f)
    def transactional_f(service, *args, **kwargs):
        session_factory = _get_session_factory()
        trans_close = False

        if hasattr(service, _TX_HOLDER_ATTRIBUTE):
            tx_holder = getattr(service, _TX_HOLDER_ATTRIBUTE)
            assert tx_holder.has_session()
        else:
            tx_holder = _get_tx_holder_factory().create_transaction_holder()
            setattr(service, _TX_HOLDER_ATTRIBUTE, tx_holder)
            trans_close = True

        try:
            logging.getLogger(__name__).debug("Entering transactional "
                                              "method.")
            if not tx_holder.has_session():
                session = session_factory.create_session()
                tx_holder.inject_session(session)

            rv = f(service, *args, **kwargs)

            if trans_close:
                tx_holder.run_transaction_callbacks()
                tx_holder.commit()

            return rv
        except:
            if trans_close:
                tx_holder.rollback()
            raise
        finally:
            logging.getLogger(__name__).debug("Finishing transactional "
                                              "method.")
            if trans_close:
                tx_holder.close()
                delattr(service, _TX_HOLDER_ATTRIBUTE)
                logging.getLogger(__name__).debug("Closing TX session.")

    return transactional_f



"""
:mod:`zsl.db.model.app_model`
-----------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
# noinspection PyCompatibility
from builtins import str
from builtins import int
from builtins import map
from builtins import object
from future.utils import viewitems

from datetime import datetime, date
from zsl.utils.date_helper import format_date_portable, format_datetime_portable

DATE_DATA = 'date_data'
DATETIME_DATA = 'datetime_data'
RELATED_FIELDS = 'related_fields'
RELATED_FIELDS_CLASS = 'cls'
RELATED_FIELDS_HINTS = 'hints'

ISO_8601_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


class AppModel(object):
    """AppModel's are used as a thin and simple communication objects. Also
    they can be saved into cache.

    .. automethod:: __init__
    """

    _not_serialized_attributes = ['_not_serialized_attributes', '_hints', '_id_name']

    def __init__(self, raw, id_name='id', hints=None):
        """
        The application model model constructor.

        :param raw: Dictionary of properties of the raw data.
        :param id_name: Name of the identifier property.
        :param hints: Tells which of the raw attributes are date or datetime string and what is theirs format \
        Example: ``{DATE_DATA: {'birthday': '%d.%m.%Y'}, DATETIME_DATA: {'created': '%Y-%m-%d %H:%M:%S' }}`` \
        this attributes are then saved in the standard zsl service date/datetime format (consult \
        zsl.utils.date_helper)
        """

        if hints is None:
            self._hints = {DATE_DATA: {}, DATETIME_DATA: {}, RELATED_FIELDS: {}}
        else:
            self._hints = hints
            if DATE_DATA not in self._hints:
                self._hints[DATE_DATA] = {}
            if DATETIME_DATA not in self._hints:
                self._hints[DATETIME_DATA] = {}
            if RELATED_FIELDS not in self._hints:
                self._hints[RELATED_FIELDS] = {}

        self.set_from_raw_data(raw)
        self._id_name = id_name

    def set_from_raw_data(self, raw):
        for k, v in viewitems(raw):
            if isinstance(v, (type(None), str, int, float, bool)):
                if k in self._hints[DATE_DATA] and v is not None:
                    d = datetime.strptime(v, self._hints[DATE_DATA][k]).date()
                    setattr(self, k, format_date_portable(d))
                elif k in self._hints[DATETIME_DATA] and v is not None:
                    d = datetime.strptime(v, self._hints[DATETIME_DATA][k])
                    setattr(self, k, format_datetime_portable(d))
                else:
                    setattr(self, k, v)
            elif isinstance(v, datetime):
                setattr(self, k, format_datetime_portable(v))
            elif isinstance(v, date):
                setattr(self, k, format_date_portable(v))
            elif k in self._hints[RELATED_FIELDS]:
                related_cls = self._hints[RELATED_FIELDS][k][RELATED_FIELDS_CLASS]
                related_hints = self._hints[RELATED_FIELDS][k].get(RELATED_FIELDS_HINTS)
                if isinstance(v, (list, tuple)):
                    setattr(self, k, [related_cls(x, 'id', related_hints) for x in v])
                else:
                    setattr(self, k, related_cls(v.__dict__, 'id', related_hints))

    def get_id(self):
        return self.__dict__[self._id_name]

    def _set_id_name(self, id_name):
        self._id_name = id_name

    @staticmethod
    def convert(v):
        if isinstance(v, AppModel):
            return v.get_attributes()
        else:
            return v

    def get_attributes(self):
        d = dict(self.__dict__)

        for k in self.__dict__:
            if k in self._not_serialized_attributes:
                d.pop(k)

            elif isinstance(d[k], AppModel):
                d[k] = self.convert(d[k])

            elif isinstance(d[k], list):
                d[k] = list(map(self.convert, d[k]))

            elif isinstance(d[k], tuple):
                d[k] = list(map(self.convert, d[k]))
                d[k] = tuple(d[k])

            elif isinstance(d[k], dict):
                for key, value in viewitems(getattr(self, k)):
                    d[k][key] = self.convert(value)

        return d

    def __str__(self):
        return "{0}: {1}".format(self.__class__, self.__dict__)



from __future__ import unicode_literals
import types


def app_model(model):
    """
    Just a convenience shortcut for `model.get_app_model()`
    """

    return model.get_app_model()


# TODO: This is quite hidden, and we could replace the function to a more convenient location.
def app_models(raw_models):
    """
    For a list of raw models return the list of the corresponding app models.
    """
    return [model.get_app_model() if model is not None else None for model in raw_models]


def app_model_or_none(raw_model):
    """
    Transforms `raw_model` to its application model. Function can handle `None` value.
    """
    return raw_model.get_app_model() if raw_model is not None else None


def visit_app_model(model, visitor):
    original_get_app_model = model.get_app_model

    def wrap(self):
        app_model = original_get_app_model()
        visitor(app_model, model)

        return app_model

    model.get_app_model = types.MethodType(wrap, model)

    return model



"""
:mod:`zsl.application.modules.cache_module`
-------------------------------------------
"""
from __future__ import unicode_literals

import logging
from injector import singleton, Module, Binder

from zsl.cache.redis_cache_module import RedisCacheModule
from zsl.cache.cache_module import CacheModule
from zsl.cache.redis_id_helper import RedisIdHelper
from zsl.cache.id_helper import IdHelper


class RedisCacheInjectionModule(Module):
    """Adds cache modules into to current configuration using reddis as
    backend.
    """

    def configure(self, binder):
        # type: (Binder) -> None
        """Initializer of the cache - creates the Redis cache module as the
        default cache infrastructure. The module is bound to `RedisCacheModule`
        and `CacheModule` keys. The initializer also creates `RedisIdHelper`
        and bounds it to `RedisIdHelper` and `IdHelper` keys.

        :param Binder binder: The binder object holding the binding context, we\
         add cache to the binder.
        """
        redis_cache_module = RedisCacheModule()
        binder.bind(
            RedisCacheModule,
            to=redis_cache_module,
            scope=singleton
        )
        binder.bind(
            CacheModule,
            to=redis_cache_module,
            scope=singleton
        )

        redis_id_helper = RedisIdHelper()
        binder.bind(
            RedisIdHelper,
            to=redis_id_helper,
            scope=singleton
        )
        binder.bind(
            IdHelper,
            to=redis_id_helper,
            scope=singleton
        )

        logging.debug("Created RedisCache binding.")



"""
:mod:`zsl.application.initializers.web_initializer`
---------------------------------------------------
"""
from __future__ import unicode_literals

from zsl.interface.webservice.web_application_loader import load


class WebInitializer(object):
    """Initialize the web application."""
    @staticmethod
    def initialize():
        load()



"""
:mod:`zsl.interface.task`
-------------------------

.. moduleauthor:: Peter Morihladko <morihladko@atteq.com>,
                  Martin Babka <babka@atteq.com>
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *

from functools import wraps
from json.encoder import JSONEncoder
from typing import Callable

import sys
from future.builtins import str

import json

from zsl import inject
from zsl.errors import ZslError
from zsl.router.task import TaskRouter
from zsl.task.job_context import JobContext, Job, DelegatingJobContext
from zsl.task.task_data import TaskData
from zsl.utils.reflection_helper import is_scalar


class ModelConversionError(Exception):
    pass


def payload_into_model(model_type, argument_name='request', remove_data=True):
    # type: (Callable, str)->Callable
    def wrapper(f):
        def fill_recursively(data, obj):
            for k, v in data.items():
                if not hasattr(obj, k):
                    raise ModelConversionError()

                if is_scalar(v):
                    setattr(obj, k, v)
                else:
                    fill_recursively(v, getattr(obj, k))

        @wraps(f)
        def executor(*args, **kwargs):
            data = args[1]  # type: TaskData
            model = model_type()
            fill_recursively(data.payload, model)
            kwargs[argument_name] = model
            if remove_data:
                args = args[:-1]
            return f(*args, **kwargs)

        return executor

    return wrapper


def create_simple_model(name, items, defaults=None, parent=object,
                        model_module=None):
    if defaults is None:
        defaults = {}
    default_code = "self.{0} = {0} if {0} is not None " \
                   "else (defaults.get('{0}', None))"
    item_definitions = "; ".join(map(lambda i: default_code.format(i), items))
    arglist = '=None, '.join(items) + ("=None, **kwargs" if len(items) else "**kwargs")
    class_code = """
class {name}(parent):
    def __new__(_cls, {arglist}):
        self = parent.__new__(_cls, **kwargs)
        {items}     
        return self
    """.format(name=name, items=item_definitions, arglist=arglist)
    namespace = {'parent': parent, 'defaults': defaults}
    exec(class_code, namespace)
    result = namespace[name]
    result._source = class_code
    if model_module is None:
        try:
            model_module = sys._getframe(1).f_globals.get('__name__',
                                                          '__main__')
        except (AttributeError, ValueError):
            pass
    if model_module is not None:
        result.__module__ = model_module
    return result


class RequestJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, object):
            return dict(o.__dict__)
        else:
            return JSONEncoder.default(self, o)


def exec_task(task_path, data):
    """Execute task.

    :param task_path: task path
    :type task_path: str|Callable
    :param data: task's data
    :type data: Any
    :return:
    """
    if not data:
        data = {'data': None, 'path': task_path}

    elif not isinstance(data, (str, bytes)):
        data = {'data': json.dumps(data, cls=RequestJSONEncoder),
                'path': task_path}

    # Open the data from file, if necessary.
    elif data is not None and data.startswith("file://"):
        with open(data[len("file://"):]) as f:
            data = f.read()

    # Prepare the task.
    job = Job(data)
    (task, task_callable) = create_task(task_path)

    # Set the context for task execution.
    old_jc = JobContext.get_current_context()
    jc = DelegatingJobContext(job, task, task_callable, old_jc)
    JobContext.set_current_context(jc)

    # Run the task.
    try:
        return jc.task_callable(jc.task_data)
    finally:
        # Set the current context back.
        JobContext.set_current_context(old_jc)


@inject(task_router=TaskRouter)
def create_task(task_path, task_router):
    if isinstance(task_path, str):
        (task, task_callable) = task_router.route(task_path)
    elif task_path is Callable or isinstance(task_path, type):
        task = task_path()
        task_callable = task.perform
    else:
        raise ZslError(
            "Can not create task with path '{0}'.".format(task_path))

    return task, task_callable



"""
Created on May 14, 2014

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from unittest.case import TestCase

from injector import inject

from zsl import Zsl, inject
from zsl.testing.db import DbTestCase

from sqlalchemy.ext.declarative import declarative_base

from zsl.resource.model_resource import ModelResource, ReadOnlyResourceUpdateOperationException, ReadOnlyResourceMixin
from resource.resource_test_helper import create_resource_test_data, UserModel

Base = declarative_base()


class ReadOnlyModelResource(ReadOnlyResourceMixin, ModelResource):

    pass


class TestReadOnlyModelResource(TestCase, DbTestCase):

    @classmethod
    def setUpClass(cls):
        create_resource_test_data()

    @inject(app=Zsl)
    def setUp(self, app):
        self.resource = app._injector.create_object(ReadOnlyModelResource, {'model_cls': UserModel})

    def testRead(self):
        m = self.resource.read([9])
        self.assertEqual('nine', m.val, "Read one")
        self.assertEqual(10, len(self.resource.read([])), 'Expecting 10 dummy models.')

    def testCreate(self):
        self.assertRaises(ReadOnlyResourceUpdateOperationException, lambda: self.resource.create({}, {}, {}))

    def testDelete(self):
        self.assertRaises(ReadOnlyResourceUpdateOperationException, lambda: self.resource.delete({}, {}, {}))

    def testUpdate(self):
        self.assertRaises(ReadOnlyResourceUpdateOperationException, lambda: self.resource.update({}, {}, {}))



"""
:mod:`zsl.utils.redis_helper`
-----------------------------

Helper module for working with redis.

.. moduleauthor::  Peter Morihladko
"""
from __future__ import unicode_literals
from future.utils import viewitems
from builtins import object
from functools import partial
from zsl import inject
from flask.config import Config


def redis_key(*args):
    """Create a string key from a hierarchical structure.

    :param args: list of keys to denote hierarchy
    :type args: list(str)
    :return: string representation of args
    :rtype: str

    :Example:
        >>> redis_key(['project', 'module', 'class'])
        'project:module:class'
    """
    # TODO why there's `is not None`
    return ':'.join(str(a) for a in args if a is not None)


class Keymaker(object):
    """Keymaker is a class to generate an object to generate Redis keys.

    :Example:
        >>> article_keys = Keymaker(prefix='articles', keys={'full_article': 'full', 'short_article': 'short'})
        >>> article_keys.full_article('today', 'ID214')
        '$PROJECT_PREFIX:articles:full:today:ID214'
    """

    # TODO I think this should be done in proper OOP

    @inject(config=Config)
    def __init__(self, keys, prefix=None, config=None):
        project_specific_prefix = config.get('REDIS', {}).get('prefix')

        for method, key in viewitems(keys):
            setattr(self, method, partial(redis_key, project_specific_prefix, prefix, key))



"""
:mod:`zsl.application.initializers.unittest_initializer`
--------------------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object

from zsl import ApplicationContext, Config, inject


class UnitTestInitializer(object):
    """Initializer handling the unit test settings."""
    @staticmethod
    @inject(config=Config)
    def initialize(config):

        if not UnitTestInitializer.is_unit_testing():
            return

        if 'TEST_DATABASE_URI' in config:
            config['DATABASE_URI'] = config['TEST_DATABASE_URI']

        if 'TEST_DATABASE_ENGINE_PROPS' in config:
            config['DATABASE_ENGINE_PROPS'] = config[
                'TEST_DATABASE_ENGINE_PROPS']

    @staticmethod
    @inject(ctx=ApplicationContext)
    def is_unit_testing(ctx):
        return ctx.unit_testing




"""
:mod:`zsl.interface.webservice.performers.task`
-----------------------------------------------
"""
from __future__ import unicode_literals
from builtins import *

import logging

from flask import request

from zsl import inject, Zsl, Injected
from zsl.interface.webservice.utils.execution import execute_web_task
from zsl.interface.webservice.utils.request_data import extract_data
from zsl.task.job_context import WebJobContext
from zsl.router.task import TaskRouter


@inject(app=Zsl)
def create_web_task(app):
    @app.route("/task/<path:path>", methods=["POST", "GET", "OPTIONS"])
    @inject(task_router=TaskRouter)
    def perform_web_task(path, task_router=Injected):
        logging.getLogger(__name__).debug("Performing task %s.", path)
        (task, task_callable) = task_router.route(path)
        jc = WebJobContext(path, extract_data(request), task, task_callable, request)
        return execute_web_task(jc, lambda: task_callable(jc.task_data))



from __future__ import absolute_import

import warnings
import functools


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1
        )

        return func(*args, **kwargs)
    return decorated



"""
:mod:`tests.resource.json_server_model_resource`
------------------------------------------------
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from typing import Optional

from future.utils import viewitems
from builtins import *

import http.client
import json
import math
import re
import urllib
from typing import Dict, List, Union, AnyStr
from unittest.case import TestCase
from resource.resource_test_helper import (
    UserModel, UserTuple, create_resource_test_data, users, get_non_existent_id, addresses
)
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl import Zsl
from zsl.application.containers.web_container import WebContainer
from zsl.resource.json_server_resource import JsonServerResource
from zsl.testing.test_utils import parent_module
from zsl.testing.http import json_loads, HTTPTestCase

# py23
if hasattr(urllib, 'parse'):
    urlencode = urllib.parse.urlencode
else:
    urlencode = urllib.urlencode


def parse_links_header(links_str):
    # type: (AnyStr) -> Dict[AnyStr, AnyStr]
    """Parse the link http header including resource metadata.

    :param links_str: link http header
    :type links_str: str
    :return: dictionary where key is the link name and value is the link
    :rtype dict[str, str]
    """

    result = {}
    parts = links_str.split(',')

    for p in parts:
        name = re.search(r'rel="([^"]+)"', p).group(1)
        url = re.search(r'<([^>]+)>', p).group(1)

        result[name] = url

    return result


def parse_page_link_headers(links_str):
    # type: (AnyStr) -> Dict[AnyStr, int]
    """Parse the link http header and get the page position from the links

    :param links_str: link http header
    :type links_str: str
    :return: dictionary where key is the link name and value is the page number
    :rtype: dict[str, int]
    """

    links = parse_links_header(links_str)

    return {
        name: int(re.search(r'_page=([\d]+)', url).group(1))
        for name, url in viewitems(links)
    }


def is_ascending(list_):
    # type: (List[int]) -> bool
    """Test if the list is in an ascending order."""

    try:
        prev = list_[0]
    except IndexError:
        return True

    for i in list_[1:]:
        if prev > i:
            return False
        prev = i
    return True


def is_descending(list_):
    # type: (List[int]) -> bool
    """Test if the list is in an descending order."""

    try:
        prev = list_[0]
    except IndexError:
        return True

    for i in list_[1:]:
        if prev < i:
            return False
        prev = i
    return True


def in_user_address(user_id, q):
    # type: (int, AnyStr) -> bool
    """Test if string is in any address property."""
    return any(a for a in addresses if a.user_id == user_id and (q in str(
        a.id) or q in a.email_address))


def in_user(user, q):
    # type: (UserTuple, AnyStr) -> bool
    """Test if string is in any user property and in its address properties."""
    return q in str(user.id) or q in user.name or in_user_address(user.id, q)


class JsonServerModelResourceTestResource(JsonServerResource):
    """Test resource based on JsonServerResource."""
    __model__ = UserModel


class JsonServerModelResourceTestCase(TestCase, HTTPTestCase):
    PATH = '/resource/json_server_model_resource_test'

    def setUp(self):
        config_object = IN_MEMORY_DB_SETTINGS.copy()
        # add this package as resource package for zsl to find the
        # `JsonServerModelResourceResource`
        config_object['RESOURCE_PACKAGES'] = ('resource',)
        zsl = Zsl(__name__, config_object=config_object,
                  modules=WebContainer.modules())
        zsl.testing = True

        # mock http requests
        self.app = zsl.test_client()

        create_resource_test_data()

    def _path(self, res_id=None, args=None, raw_args=''):
        # type: (Optional[Union[AnyStr, int]], Dict[AnyStr, Union[str, int]], AnyStr) -> str
        """Generate a resource link

        :param res_id: resource id
        :param args: optional url arguments as {name: value}
        :param raw_args: url arguments in string, appended after args
        :return: url path for resource
        """
        if args:
            args = '?' + urlencode(args)
        else:
            args = ''

        if args:
            args += '&' + raw_args
        elif raw_args:
            args = '?' + raw_args

        if res_id:
            return "{path}/{id_}{args}".format(
                path=self.PATH,
                id_=res_id,
                args=args
            )
        else:
            return "{path}{args}".format(
                path=self.PATH,
                args=args
            )

    def testCreate(self):
        user_data = {'name': 'eleven'}

        rv = self.app.post(self.PATH, data=json.dumps(user_data), content_type='application/json')
        user = json_loads(rv.data)

        self.assertEqual(user_data['name'], user['name'])
        self.assertHTTPStatus(http.client.CREATED, rv.status_code, 'http status should be 201')

        rv = self.app.get(rv.location)
        self.assertDictEqual(user, json_loads(rv.data), 'object should be saved')

    def testReadOne(self):
        rv = self.app.get(self._path(res_id=1))

        self.assertDictEqual(users[0]._asdict(), json_loads(rv.data), 'json should equal model value')
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')

    def testReadNotFound(self):
        rv = self.app.get(self._path(res_id=get_non_existent_id()))

        self.assertDictEqual({}, json_loads(rv.data), 'should return empty json')
        self.assertHTTPStatus(http.client.NOT_FOUND, rv.status_code, 'http status should be 404')

    def testUpdateWithPut(self):
        """Put - same as patch."""
        update = {'name': 'eleven'}

        original = users[0]

        rv = self.app.put(self._path(res_id=original.id), data=json.dumps(update), content_type='application/json')
        user = json_loads(rv.data)

        updated = original._asdict()
        updated['name'] = update['name']

        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertDictEqual(updated, user, 'updated value should be returned')

        rv = self.app.get(self._path(res_id=original.id))
        self.assertDictEqual(updated, json_loads(rv.data), 'updated value should be stored')

    def testUpdateWithPatch(self):
        """Patch - update partially."""

        original = users[0]
        update = {'name': 'new value'}

        rv = self.app.patch(self._path(res_id=original.id), data=json.dumps(update), content_type='application/json')
        user = json_loads(rv.data)

        updated = original._asdict()
        updated['name'] = update['name']

        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertDictEqual(updated, user, 'updated value should be returned')

        rv = self.app.get(self._path(res_id=original.id))
        self.assertDictEqual(updated, json_loads(rv.data), 'updated value should be stored')

    def testUpdateWithPatchNotFound(self):
        update = {'name': 'new value'}

        rv = self.app.patch(self._path(res_id=get_non_existent_id()), data=json.dumps(update),
                            content_type='application/json')

        self.assertDictEqual({}, json_loads(rv.data), 'should return empty json')
        self.assertHTTPStatus(http.client.NOT_FOUND, rv.status_code, 'http status should be 404')

    def testDelete(self):
        to_delete = users[0]

        rv = self.app.delete(self._path(res_id=to_delete.id))

        self.assertDictEqual({}, json_loads(rv.data), 'should return empty json')
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')

        rv = self.app.get(self._path(res_id=to_delete.id))
        self.assertDictEqual({}, json_loads(rv.data), 'should be deleted')
        self.assertHTTPStatus(http.client.NOT_FOUND, rv.status_code, 'http status should be 404')

    def testDeleteAll(self):
        rv = self.app.delete(self.PATH)

        self.assertDictEqual({}, json_loads(rv.data), 'should be empty')
        self.assertHTTPStatus(http.client.NOT_FOUND, rv.status_code, 'http status should be 404')

    def testDeleteNotFound(self):
        rv = self.app.delete(self._path(res_id=get_non_existent_id()))

        self.assertDictEqual({}, json_loads(rv.data), 'should be empty')
        self.assertHTTPStatus(http.client.NOT_FOUND, rv.status_code, 'http status should be 404')

    def testReadMany(self):
        rv = self.app.get(self.PATH)
        data = json_loads(rv.data)

        self.assertEqual(len(users), len(data), 'should return all elements')
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')

    def testReadManyWithPageAndDefaultLimit(self):
        page = 1
        default_limit = 10

        rv = self.app.get(self._path(args={'_page': page}))
        data = json_loads(rv.data)

        self.assertEqual(default_limit, len(data), 'should return 10 items')
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

    def testReadManyWithLimit(self):
        limit = 5

        rv = self.app.get(self._path(args={'_limit': limit}))
        data = json_loads(rv.data)

        self.assertEqual(limit, len(data), 'should return %d items' % limit)
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

        count = 3
        offset = len(users) - count
        limit = 5

        rv = self.app.get(self._path(args={'_limit': limit, '_start': offset}))
        data = json_loads(rv.data)

        self.assertEqual(count, len(data), 'should return %d items' % count)
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

    def testReadManyWithPage(self):
        page = 2
        limit = 3

        rv = self.app.get(self._path(args={'_page': page, '_limit': limit}))
        data = json_loads(rv.data)

        self.assertEqual(limit, len(data), 'should return %d items' % limit)
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

        pages = parse_page_link_headers(rv.headers['Links'])
        self.assertEqual(1, pages.get('first'), 'there should be link to first page')
        self.assertEqual(page - 1, pages.get('prev'), 'there should be link to prev page')
        self.assertEqual(page + 1, pages.get('next'), 'there should be link to next page')
        self.assertEqual(math.ceil(len(users) / limit), pages.get('last'),
                         'there should be link to last page')

    def testReadManyWithStartAndEnd(self):
        end = 7

        rv = self.app.get(self._path(args={'_end': end}))
        data = json_loads(rv.data)

        self.assertEqual(end, len(data), 'should return %d items' % end)
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

        start = 3
        end = 10

        rv = self.app.get(self._path(args={'_start': start, '_end': end}))
        data = json_loads(rv.data)

        self.assertEqual(end - start, len(data), 'should return %d items' % (end - start))
        self.assertHTTPStatus(http.client.OK, rv.status_code, 'http status should be 200')
        self.assertEqual(len(users), int(rv.headers['X-Total-Count']),
                         'response header \'X-Total-Count\' should be set')

    def testReadManyWithSort(self):
        rv = self.app.get(self._path(args={'_sort': 'id'}))
        data = json_loads(rv.data)

        self.assertTrue(is_ascending([d['id'] for d in data]), 'default sort direction should be ascend')

        rv = self.app.get(self._path(args={'_sort': 'id', '_order': 'DESC'}))
        data = json_loads(rv.data)

        self.assertTrue(is_descending([d['id'] for d in data]), 'sort direction should be descent')

    def testReadManyWithFilters(self):
        rv = self.app.get(self._path(args={'name': 'nine'}))
        data = json_loads(rv.data)

        expected_length = len([d for d in users if d.name == 'nine'])
        self.assertEqual(expected_length, len(data), 'should filter with help of \'name\' param')

        rv = self.app.get(self._path(raw_args='name=nine&name=seven'))
        data = json_loads(rv.data)

        expected_length = len([d for d in users if d.name == 'nine' or d.name == 'seven'])
        self.assertEqual(expected_length, len(data),
                         'should filter with help of OR operator and supporting multiple key values')

        rv = self.app.get(self._path(raw_args='id_gte={id1}&name={name1}&id_lte={id2}&name_like={name2}'.format(
            id1=len(users), id2=1, name1='seven', name2='eig'
        )))
        data = json_loads(rv.data)

        expected_length = len([d for d in users
                               if d.id <= 1 or d.id >= len(users) or d.name == 'seven' or 'eig' in d.name])
        self.assertEqual(expected_length, len(data),
                         'should filter with help of various operators')

        rv = self.app.get(self._path(args={'not_existent_column': 'x', 'id': 1}))
        data = json_loads(rv.data)

        self.assertEqual(1, len(data), 'should ignore unknown column')

    def testFulltextQuery(self):
        q = '@'

        rv = self.app.get(self._path(args={'q': q}))
        data = json_loads(rv.data)

        expected_length = len([u for u in users if in_user(u, q)])
        self.assertEqual(expected_length, len(data), 'should filter using a fulltext query')

    def testEmbed(self):
        user_id = 1
        rv = self.app.get(self._path(res_id=user_id, args={'_embed': 'addresses'}))
        data = json_loads(rv.data)

        expected_length = len([a for a in addresses if a.user_id == user_id])
        self.assertEqual(expected_length, len(data['addresses']), 'should return user with embedded addresses')

        rv = self.app.get(self._path(args={'_embed': 'addresses', '_sort': 'id', '_page': 1}))
        data = json_loads(rv.data)

        for u in data:
            expected_length = len([a for a in addresses if a.user_id == u['id']])
            self.assertEqual(expected_length, len(u['addresses']), 'should return users with embedded addresses')



"""
Test resource guard.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from unittest import TestCase
import inspect

try:
    import unittest.mock as mock
except ImportError:
    import mock

from zsl.resource.guard import guard, GuardedMixin, ResourcePolicy, Access

_methods = ['create', 'read', 'update', 'delete']

TEST_VALUE_CREATED = 'created'
TEST_VALUE_READ = 'read'
TEST_VALUE_UPDATED = 'updated'
TEST_VALUE_DELETED = 'deleted'


class AccessError(Exception):
    """Because this is not a subtype of PolicyViolationError it should be not
    caught by the default exception handler, so we can test it with an assert.
    """
    pass


class TestResourceGuard(TestCase):
    def setUp(self):
        class PlainResource(object):
            def create(self):
                return TEST_VALUE_CREATED

            def read(self):
                return TEST_VALUE_READ

            def update(self):
                return TEST_VALUE_UPDATED

            def delete(self):
                return TEST_VALUE_DELETED

        class ToBeSecuredResource(PlainResource, GuardedMixin):
            pass

        self.PlainResource = PlainResource
        self.ToBeSecuredResource = ToBeSecuredResource

    @staticmethod
    def _createMockPolicy(before_side_effect=None, after_side_effect=None):
        policy = ResourcePolicy()

        for m in _methods:
            before = 'can_%s__before' % m
            after = 'can_%s__after' % m
            setattr(policy, before, mock.Mock(
                return_value=Access.ALLOW,
                side_effect=before_side_effect,
                name=before
            ))
            setattr(policy, after, mock.Mock(
                return_value=Access.ALLOW,
                side_effect=after_side_effect,
                name=after
            ))

        return policy

    def testAddGuardedMixin(self):
        test_resource_cls = guard([])(self.PlainResource)

        self.assertTrue(
            issubclass(test_resource_cls, GuardedMixin),
            'should create a new class with GuardedMixin in its base'
        )

        self.assertNotEqual(self.PlainResource, test_resource_cls,
                            'should create a new class')

    def testForGuardedMethods(self):
        test_resource_cls = guard([])(self.PlainResource)
        test_resource = test_resource_cls()

        for method in ['create', 'read', 'update', 'delete']:
            m_name = 'guarded_%s' % method
            self.assertTrue(hasattr(test_resource, m_name) and
                            inspect.ismethod(getattr(test_resource, m_name)),
                            'should have %s method' % m_name)

    def testDontModifyTheClassWhenTheMixinIsPresent(self):
        test_resource_cls = guard([])(self.ToBeSecuredResource)

        self.assertEqual(self.ToBeSecuredResource, test_resource_cls,
                         'should not create a new class when GuardedMixin is in'
                         ' the bases')

    def testShouldAllowAccess(self):
        class PermissivePolicy(ResourcePolicy):
            default = Access.ALLOW

        policy = PermissivePolicy()
        test_resource_cls = guard([policy])(self.ToBeSecuredResource)
        resource = test_resource_cls()

        self.assertEqual(TEST_VALUE_CREATED, resource.create(),
                         'should execute create and return value')
        self.assertEqual(TEST_VALUE_READ, resource.read(),
                         'should execute read and return value')
        self.assertEqual(TEST_VALUE_UPDATED, resource.update(),
                         'should execute update and return value')
        self.assertEqual(TEST_VALUE_DELETED, resource.delete(),
                         'should execute delete and return value')

    def testApplyTwice(self):
        policy_1 = [1, 2, 3]
        policy_2 = [4, 5, 6]
        all_policies = policy_2 + policy_1

        test_resource_cls = guard(policy_2)(
            guard(policy_1)(
                self.ToBeSecuredResource
            )
        )

        self.assertListEqual(all_policies, test_resource_cls._guard_policies,
                             'should join guard policies')

    def testDenyWithException(self):
        class ForbiddenPolicy(ResourcePolicy):
            @property
            def default(self):
                raise AccessError

        policy = ForbiddenPolicy()
        test_resource_cls = guard([policy])(self.ToBeSecuredResource)
        resource = test_resource_cls()

        with self.assertRaises(AccessError):
            resource.create()

        with self.assertRaises(AccessError):
            resource.read()

        with self.assertRaises(AccessError):
            resource.update()

        with self.assertRaises(AccessError):
            resource.delete()

    def testErrorHandler(self):
        class ForbiddenPolicy(ResourcePolicy):
            default = Access.DENY

        error_result = 'handled'

        def exc_handler(*_):
            return error_result

        test_resource_cls = guard(
            policies=[ForbiddenPolicy()],
            exception_handlers=[exc_handler]
        )(self.ToBeSecuredResource)
        resource = test_resource_cls()

        self.assertEqual(error_result, resource.create(),
                         'should execute the handler and return an error value')
        self.assertEqual(error_result, resource.read(),
                         'should execute the handler and return an error value')
        self.assertEqual(error_result, resource.update(),
                         'should execute the handler and return an error value')
        self.assertEqual(error_result, resource.delete(),
                         'should execute the handler and return an error value')

    def testCallAfterAndBeforeCallbacks(self):
        policy = self._createMockPolicy()

        class TestResource(self.ToBeSecuredResource):
            def create(self):
                policy.can_create__before.assert_called()
                policy.can_create__after.assert_not_called()

            def read(self):
                policy.can_read__before.assert_called()
                policy.can_read__after.assert_not_called()

            def update(self):
                policy.can_update__before.assert_called()
                policy.can_update__after.assert_not_called()

            def delete(self):
                policy.can_delete__before.assert_called()
                policy.can_delete__after.assert_not_called()

        test_resource_cls = guard([policy])(TestResource)
        resource = test_resource_cls()

        resource.create()
        policy.can_create__before.assert_called()

        resource.read()
        policy.can_create__before.assert_called()

        resource.update()
        policy.can_create__before.assert_called()

        resource.delete()
        policy.can_delete__before.assert_called()

    def testWrappers(self):
        def transform_value(str_):
            return 'transfered_' + str_

        def wrapper(fn):
            def wrapped(*args, **kwargs):
                rv = fn(*args, **kwargs)

                return transform_value(rv)

            return wrapped

        class PermissivePolicy(ResourcePolicy):
            default = Access.ALLOW

        policy = PermissivePolicy()
        test_resource_cls = guard(
            policies=[policy],
            method_wrappers=[wrapper]
        )(self.ToBeSecuredResource)

        resource = test_resource_cls()

        self.assertEqual(transform_value(TEST_VALUE_CREATED), resource.create(),
                         'should transform the value')
        self.assertEqual(transform_value(TEST_VALUE_READ), resource.read(),
                         'should transform the value')
        self.assertEqual(transform_value(TEST_VALUE_UPDATED), resource.update(),
                         'should transform the value')
        self.assertEqual(transform_value(TEST_VALUE_DELETED), resource.delete(),
                         'should transform the value')

    def testPolicyChaining(self):
        class AllowReadPolicy(ResourcePolicy):
            can_read = Access.ALLOW

        class AllowUpdatePolicy(ResourcePolicy):
            can_update = Access.ALLOW

        class DenyDeletePolicy(ResourcePolicy):
            can_delete = Access.DENY

        error_result = 'handled'

        def exc_handler(*_):
            return error_result

        test_resource_cls = guard(
            policies=[AllowReadPolicy(), AllowUpdatePolicy(),
                      DenyDeletePolicy()],
            exception_handlers=[exc_handler]
        )(self.ToBeSecuredResource)

        resource = test_resource_cls()

        self.assertEqual(TEST_VALUE_READ, resource.read(),
                         "should allow to read, because of AllowReadPolicy")
        self.assertEqual(TEST_VALUE_UPDATED, resource.update(),
                         "should allow to update, because of AllowUpdatePolicy")
        self.assertEqual(error_result, resource.create(),
                         "should deny to create, because there\'s no policy "
                         "to allow it")
        self.assertEqual(error_result, resource.delete(),
                         "should deny to delete, because of DenyDeletePolicy")




"""
:mod:`zsl.utils.resource_helper`
--------------------------------

Helper module for resource management.
"""
# TODO describe what model resource is and use cases

from __future__ import unicode_literals
import importlib
import logging
from typing import Callable, Union, List, Dict
from flask import request

from zsl.utils.string_helper import underscore_to_camelcase
from zsl.utils.task_helper import instantiate
from zsl.resource.model_resource import ModelResource
from zsl.interface.resource import ResourceResult

from zsl import inject, Config, Injected, Zsl


class MethodNotImplementedException(Exception):
    """Exception raised on missing method"""
    pass


def parse_resource_path(path):
    """Split the path to its elements.

    :param path: URL path
    :type path: str
    :return: name and rest of the path
    :rtype: tuple(str, list(str))
    """
    splits = path.split('/')

    return splits[0], splits[1:]


def get_method(resource, method):
    """Test and return a method by name on resource.

    :param resource: resource object
    :type resource: object
    :param method: method name
    :type method: str
    :return: bounded method
    :raises MethodNotImplementedException: when method not found
    """
    if hasattr(resource, method) and callable(getattr(resource, method)):
        return getattr(resource, method)
    else:
        raise MethodNotImplementedException()


@inject(config=Config)
def get_resource_task(resource_path, config=Injected):
    # type: (str, Config) -> Callable[[str, Dict, Dict], Union[List[AppModel], AppModel, ResourceResult]]
    """Search and return a bounded method for given path.

    :param resource_path: resource path
    :type resource_path: str
    :param config: current configuration, injected
    :type config: Config
    :return: bounded method or None when not found
    :raises NameError: when can't find given resource by path
    """
    class_name = underscore_to_camelcase(resource_path) + 'Resource'

    resource_packages = config['RESOURCE_PACKAGES'] if 'RESOURCE_PACKAGES' in config else (config['RESOURCE_PACKAGE'],)

    resource = None
    for resource_package in resource_packages:
        module_name = "{0}.{1}".format(resource_package, resource_path)

        try:
            module = importlib.import_module(module_name)

            try:
                cls = getattr(module, class_name)

                resource = instantiate(cls)
                break
            except AttributeError:
                raise NameError("Can't find resource [{0}]".format(resource_path))

        except ImportError:

            # try to find __exposer__
            try:
                exposer = importlib.import_module(resource_package + '.__exposer__')

                resource = exposer.get_resource(class_name)
                break
            except Exception as e:
                logging.error("Can not load resource {0} [{1}].".format(resource_path, e))
                pass

    if resource is None:
        raise NameError("Can't find resource [{0}]".format(resource_path))

    try:
        if request.method == 'POST':
            return get_method(resource, 'create')

        elif request.method == 'GET':
            return get_method(resource, 'read')

        elif request.method == 'PUT' or request.method == 'PATCH':
            return get_method(resource, 'update')

        elif request.method == 'DELETE':
            return get_method(resource, 'delete')

        else:
            logging.error("Invalid request method [%s] is requested for path [%s]", request.method, resource_path)

    except MethodNotImplementedException:
        logging.error(
            "MethodNotImplementedException raised for method [%s] and path [%s]",
            request.method,
            resource_path
        )
        return None


@inject(app=Zsl)
def create_model_resource(resource_map, name, app=Injected):
    """Create a model resource from a dict ``resource_map``
    {'resource name': ('model package', 'model class')}

    :param resource_map: dict with resource descriptions
    :type resource_map: dict(str, tuple(str))
    :param name: name of the concrete resource
    :param app: current application, injected
    :type app: Zsl
    :return: instantiated model resource
    """
    try:
        resource_description = resource_map[name]
        if len(resource_description) == 2:
            module_name, model_name = resource_map[name]
            resource_class = ModelResource
        elif len(resource_description) == 3:
            module_name, model_name, resource_class = resource_map[name]
        else:
            raise ImportError("Invalid resource description for resource '{0}'".format(name))
    except KeyError:
        raise ImportError("Missing resource description for resource '{0}'".format(name))

    module = importlib.import_module(module_name)
    model_cls = getattr(module, model_name)

    return app.injector.create_object(resource_class, {'model_cls': model_cls})



"""
:mod:`zsl.db.model.sql_alchemy`
-------------------------------

one base the rule them all

.. moduleauthor:: Peter Morihladko
"""
from __future__ import unicode_literals

from sqlalchemy.ext.declarative import declarative_base

from zsl.db.model.raw_model import ModelBase

DeclarativeBase = declarative_base(cls=ModelBase)
metadata = DeclarativeBase.metadata



"""
:mod:`zsl.interface.resource`
-----------------------------

Helper types describing resource results for resource router.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from collections import namedtuple

result_fields = ['body', 'count', 'links', 'status', 'location', 'expose_headers']

ResourceResult = namedtuple('ResourceResult', result_fields)
ResourceResult.__new__.__defaults__ = (None,) * len(ResourceResult._fields)  # allow creating an empty result



# -*- coding: utf-8 -*-
"""
:mod:`zsl.utils.date_helper`
----------------------------
"""
from __future__ import unicode_literals
from builtins import *


def documentation_link(chapter):
    # type: (str)->str
    """
    Creates a link to the documentation.

    This method is useful for showing a link to the ZSL documentation in case of any misconfiguration, etc.
    :param chapter: Chapter name in to which the link points. Use underscores instead of spaces.
    :return: The absolute link to the documentation.
    """
    return "http://zsl.readthedocs.io/en/latest/{0}.html".format(chapter)



"""
:mod:`zsl.utils.task_helper`
----------------------------

Helper module for task management.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from zsl.task.task_data import TaskData

TASK_PERFORM_METHOD = "perform"


def run_task(task_cls, task_data):
    """Instantiate and run the perform method od given task data.

    :param task_cls: task class
    :param task_data: task data
    :type task_data: TaskData
    :return: task's result
    """
    task = instantiate(task_cls)
    task_callable = get_callable(task)
    return task_callable(TaskData(task_data))


def run_task_json(task_cls, task_data):
    """Instantiate and run the perform method od given task data.

    :param task_cls: task class
    :param task_data: task data
    :type task_data: TaskData
    :return: task's result
    """
    # TODO what does set_skipping_json do?
    task = instantiate(task_cls)
    task_callable = get_callable(task)
    td = TaskData(task_data)
    td.set_skipping_json(True)
    return task_callable(td)


def get_callable(task):
    """Return the perform method on given task.

    :param task: task instance
    :return: bound perform method
    """
    return getattr(task, TASK_PERFORM_METHOD)


from zsl.utils.injection_helper import instantiate



"""
:mod:`zsl.utils.deploy.apiary_doc_generator`
--------------------------------------------

Generates the API documentation suitable for the apiary.io. It parses all the files and finds the apiary.io definitions
in the documentary comments. Then outputs it to a file.

.. moduleauthor:: Martin Babka
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import sys
import os

import pydoc
import inspect
import pkgutil
import importlib
import logging

from zsl import inject
from zsl.router.method import get_method_packages
from zsl.router.task import TaskRouter


class ApiaryDoc(pydoc.Doc):

    _API_DOC_STR = "API Documentation:"

    def __init__(self):
        self._docs = []
        self._done = set()

    @staticmethod
    def _get_obj_id(obj):
        if hasattr(obj, '__path__'):
            return obj.__path__[0]
        if hasattr(obj, '__file__'):
            return obj.__file__
        if hasattr(obj, '__name__'):
            return obj.__name__
        else:
            return obj

    def _add_doc(self, obj):
        obj_id = self._get_obj_id(obj)
        if obj_id in self._done:
            return
        logging.debug('Adding {0}.'.format(obj_id))
        self._done.add(obj_id)

        apistr = ""
        try:
            if obj.__doc__ is None:
                return

            start = obj.__doc__.find(self._API_DOC_STR)
            if start == -1:
                return
            apistr = obj.__doc__[start + len(self._API_DOC_STR):]
        except:
            return

        apistr = apistr.splitlines()
        while apistr[0].isspace() or apistr[0] == "":
            apistr = apistr[1:]

        def white_space_at_beginning(l):
            c = 0
            i = 0
            length = len(l)
            while i < length:
                if l[i].isspace():
                    c += 1
                else:
                    return c
                i += 1
            return c

        m = white_space_at_beginning(apistr[0])
        for l in apistr:
            if l.isspace() or l == "":
                continue
            wl = white_space_at_beginning(l)
            if m > wl:
                m = wl
        apistr = [x[m:] for x in apistr]

        self._docs.append("\n".join(apistr))

    def docmodule(self, obj, name=None, *args):
        if self._get_obj_id(obj) in self._done:
            return

        self._add_doc(obj)
        for _key, value in inspect.getmembers(obj, inspect.isclass):
            self._add_doc(value)
        for _key, value in inspect.getmembers(obj, lambda x: hasattr(x, '__call__')):
            self._add_doc(value)

        for _key, value in inspect.getmembers(obj, inspect.ismodule):
            self.docmodule(value)

        if not hasattr(obj, '__path__'):
            return
        if obj.__path__[0].startswith(os.path.dirname(sys.executable)):
            return

        for loader, module_name, _ispkg in pkgutil.iter_modules(obj.__path__):
            logging.debug("Loading module {0} in {1}.".format(module_name, obj.__path__))
            try:
                module = loader.find_module(module_name).load_module(module_name)
                self.docmodule(module)
            except:
                pass

    def get_doc(self):
        return "FORMAT: 1A\n\n" + "\n\n".join(self._docs)


@inject(task_router=TaskRouter)
def generate_apiary_doc(task_router):
    """Generate apiary documentation.

    Create a Apiary generator and add application packages to it.

    :param task_router: task router, injected
    :type task_router: TaskRouter
    :return: apiary generator
    :rtype: ApiaryDoc
    """
    generator = ApiaryDoc()

    for m in task_router.get_task_packages() + get_method_packages():
        m = importlib.import_module(m)
        generator.docmodule(m)

    return generator



"""
:mod:`zsl.application.error_handler`
---------------------------------------------------

This module does the error handling. It allows users to register
an error handler for a given exception type. It also provides default
error handlers.
"""
from __future__ import unicode_literals
from builtins import *

import http.client
import logging
import traceback

from abc import abstractmethod
from flask import request
from functools import wraps

from zsl.db.model.app_model import AppModel
from zsl.router.task import RoutingError
from zsl.task.job_context import add_responder, StatusCodeResponder

from zsl.utils.documentation import documentation_link

_error_handlers = []


class ErrorResponse(AppModel):
    def __init__(self, code, message):
        super(ErrorResponse, self).__init__({})
        self.code = code
        self.message = message


class ErrorHandler(object):
    @abstractmethod
    def can_handle(self, e):
        pass

    @abstractmethod
    def handle(self, e):
        pass


def register(e):
    # type: (ErrorHandler)->None
    _error_handlers.append(e)


class DefaultErrorHandler(ErrorHandler):
    def can_handle(self, e):
        return True

    def handle(self, ex):
        logger = logging.getLogger(__name__)
        logger.error(str(ex) + "\n" + traceback.format_exc())
        logger.error("Request:\n{0}\n{1}\n".format(request.headers,
                                                   request.data))
        logger.info("Provide your own error handler so that "
                    "a better error is produced, check {0}.".format(
            documentation_link('error_handling')))

        add_responder(StatusCodeResponder(http.client.INTERNAL_SERVER_ERROR))
        return ErrorResponse('UNKNOWN_ERROR', "An error occurred!")


class RoutingErrorHandler(ErrorHandler):
    def can_handle(self, e):
        return isinstance(e, RoutingError)

    def handle(self, ie):
        logging.error(str(ie) + "\n" + traceback.format_exc())
        add_responder(StatusCodeResponder(404))
        return str(ie)


_DEFAULT_ERROR_HANDLER = DefaultErrorHandler()
_ROUTING_ERROR_HANDLER = RoutingErrorHandler()


def error_handler(f):
    """
    Default error handler.
     - On server side error shows a message
       'An error occurred!' and returns 500 status code.
     - Also serves well in the case when the resource/task/method
       is not found - returns 404 status code.
    """

    @wraps(f)
    def error_handling_function(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except RoutingError as ie:
            return _ROUTING_ERROR_HANDLER.handle(ie)
        except Exception as ex:
            for eh in _error_handlers:
                if eh.can_handle(ex):
                    return eh.handle(ex)

            return _DEFAULT_ERROR_HANDLER.handle(ex)

    return error_handling_function



"""
:mod:`zsl.db.helpers.nested`
----------------------------

.. moduleauthor:: Peter Morihladko <morihladko@atteq.com>
"""
from __future__ import unicode_literals
from builtins import *
from typing import Any
from typing import Optional

from sqlalchemy.orm.attributes import InstrumentedAttribute

from zsl.db.model.app_model import AppModel
from zsl.db.model.raw_model import ModelBase


def get_nested_field_name(field):
    if isinstance(field, InstrumentedAttribute):
        field = field.property.key

    return field


def nested_model(model, nested_fields):
    # type: (ModelBase, Any)->Optional[AppModel]
    """Return :class:`zsl.db.model.app_model import AppModel` with the nested 
    models attached. ``nested_fields`` can be a simple list as model fields, 
    or it can be a tree definition in dict with leafs as keys with None value
    """
    if model is None:
        return None

    app_model = model.get_app_model()
    is_dict = isinstance(nested_fields, dict)

    for field in nested_fields:
        field = get_nested_field_name(field)

        nested_nested = nested_fields.get(
            field) if is_dict and nested_fields.get(field) else []
        value = getattr(model, field, None)

        # we can have also lists in field
        nm_fn = nested_models if isinstance(value, list) else nested_model

        setattr(app_model, field, nm_fn(value, nested_nested))

    return app_model


def nested_models(models, nested_fields):
    """For a list of ``models`` apply ``get_nested_model`` with given 
    ``nested_fields``.
    """
    return [nested_model(model, nested_fields) for model in models]



"""
:mod:`zsl.interface.webservice.performers.method`
-------------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

import logging
from importlib import import_module

import sys

from zsl.router.method import get_method_packages


def call_exposers_in_method_packages():
    for package in get_method_packages():
        if package in sys.modules:
            module = sys.modules[package]
            if hasattr(module, '__reloader__'):
                getattr(module, '__reloader__')()
        else:
            module = import_module(package)

        msg = "Calling exposers in method package {}".format(package)
        logging.getLogger(__name__).debug(msg)
        if hasattr(module, '__exposer__'):
            getattr(module, '__exposer__')()



from functools import wraps
from typing import Any, Callable

from flask import request
from flask.helpers import make_response
from flask.wrappers import Response

from zsl import Zsl, inject
from zsl.application.error_handler import error_handler
from zsl.interface.webservice.utils.response_headers import append_headers
from zsl.task.job_context import JobContext, WebJobContext


def notify_responders(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        response = f(*args, **kwargs)
        job_context = JobContext.get_current_context()
        job_context.notify_responders(response)
        return response

    return wrapper


def create_web_response(result):
    # type:(Any,WebJobContext)->Response
    return make_response(result)


def convert_to_web_response(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)
        return create_web_response(result)

    return wrapper


@append_headers
@notify_responders
@convert_to_web_response
@error_handler
@inject(app=Zsl)
def execute_web_task(job_context, callable, app):
    # type:(WebJobContext,Callable,Zsl)->Response
    app.logger.debug("Data found '%s'.", str(job_context.task_data.payload))
    JobContext.set_current_context(job_context)
    if request.method == 'OPTIONS':
        return app.make_default_options_response()
    else:
        return callable()



"""
:mod:`zsl.utils.php_helper`
---------------------------

.. moduleauthor:: Peter Morihladko

Module with functions to help with dealing with code written in PHP.
"""
from __future__ import unicode_literals


def bool_to_str(boolean):
    """Convert ``boolean`` to string like PHP.

    :param boolean: boolean value
    :type boolean: bool
    :return: string representation
    :rtype: str
    """
    return '1' if boolean else ''



import sys
from setuptools import setup, find_packages

_is_py3 = sys.version_info > (3, 0)

requirements = [
    'flask',
    'future',
    'injector==0.12.1',
    'requests>=2.7',
    'SQLAlchemy>=1',
    'typing>=3.5',
    'Werkzeug>=0.12',
]

if sys.version_info < (3, 0):
    requirements.append('enum34')

setup(name='zsl',
      version='0.16.11',
      description='zsl application framework for web based services',
      author='Atteq s.r.o.',
      author_email='open.source@atteq.com',
      url='https://github.com/AtteqCom/zsl',
      license='MIT',
      package_dir={'': 'src'},
      install_requires=requirements,
      extras_require={
          'cli': ['bpython'],
          'redis': ['redis>=2.10'],
          'celery': ['zsl_client'],
          'gearman': ['zsl_client', 'gearman'],
          'alembic': ['alembic'],
          'documentation': ['sphinx', 'recommonmark', 'sphinx_rtd_theme',
                            'alembic']
      },
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.6',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Application Frameworks'
      ],
      packages=find_packages('src'))



"""
:mod:`zsl.utils.deploy.integrator`
----------------------------------
"""

from __future__ import unicode_literals
import tempfile
import os


def integrate_to_file(what, filename, start_line, end_line):
    """WARNING this is working every second run.. so serious bug
    Integrate content into a file withing "line marks"
    """

    try:
        with open(filename) as f:
            lines = f.readlines()
    except IOError:
        lines = []

    tmp_file = tempfile.NamedTemporaryFile(delete=False)

    lines.reverse()

    # first copy before start line
    while lines:
        line = lines.pop()

        if line == start_line:
            break

        tmp_file.write(line)

    # insert content
    tmp_file.write(start_line)
    tmp_file.write(what)
    tmp_file.write(end_line)

    # skip until end line
    while lines:
        line = lines.pop()

        if line == end_line:
            break

    # copy rest
    tmp_file.writelines(lines)
    tmp_file.close()

    os.rename(tmp_file.name, filename)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from typing import Union


def get_http_status_code_value(status_code):
    # type: (Union[int|HTTPStatus])->int
    """Py2/3 status code."""
    if hasattr(status_code, 'value'):
        return status_code.value  # python 3
    else:
        return status_code  # python 2



"""
:mod:`zsl.application.modules.celery_module`
--------------------------------------------
"""
from __future__ import unicode_literals

import click
from injector import Module, singleton

from zsl import Zsl, Injected
from zsl.application.modules.cli_module import ZslCli
from zsl.interface.task_queue import TaskQueueWorker, run_worker
from zsl.interface.celery.worker import CeleryTaskQueueWorkerBase, CeleryTaskQueueMainWorker, \
    CeleryTaskQueueOutsideWorker
from zsl.utils.injection_helper import simple_bind, inject


class CeleryTaskQueueMainWorkerModule(Module):
    """Adds celery worker to current configuration."""

    def configure(self, binder):
        worker = CeleryTaskQueueMainWorker()
        binder.bind(TaskQueueWorker, to=worker, scope=singleton)
        binder.bind(CeleryTaskQueueWorkerBase, to=worker, scope=singleton)


class CeleryTaskQueueOutsideWorkerModule(Module):
    """Adds celery outside worker to current configuration.

    Outside worker is meant to run with help of celery cli tools.
    """

    def configure(self, binder):
        worker = CeleryTaskQueueOutsideWorker()
        binder.bind(TaskQueueWorker, to=worker, scope=singleton)
        binder.bind(CeleryTaskQueueWorkerBase, to=worker, scope=singleton)


class CeleryCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) ->  CeleryCli

        @zsl_cli.cli.group(help='Celery related tasks.')
        def celery():
            pass

        @celery.command(help="run worker",
                        context_settings=dict(ignore_unknown_options=True))
        @click.argument('argv', nargs=-1, type=click.UNPROCESSED)
        @click.pass_context
        def worker(_, worker_args):
            """Run Zsl celery worker.

            :param : arguments for celery worker
            """
            run_worker(worker_args)

        self._celery = celery

    def celery(self):
        return self._celery


class CeleryCliModule(Module):
    def configure(self, binder):
        simple_bind(binder, CeleryCli, singleton)



"""
:mod:`zsl.db.model.app_model_json_encoder`
------------------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from json.encoder import JSONEncoder
from zsl.db.model import AppModel


class AppModelJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, AppModel):
            return o.get_attributes()
        elif isinstance(o, object):
            return dict(o.__dict__)
        else:
            return JSONEncoder.default(self, o)



"""
:mod:`zsl.application.initialization_context`
---------------------------------------------

The modules deals with the initialization of the basics - python path and then
loads and initializes the application if necessary.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals


class InitializationContext(object):
    def __init__(self, unit_test=False, initializers=None):
        if not initializers:
            initializers = []

        self._unit_test = unit_test
        self._initializers = initializers
        self._is_initialized = False

    def get_unit_testing(self):
        return self._unit_test

    unit_testing = property(get_unit_testing)

    @property
    def is_initialized(self):
        return self._is_initialized

    def initialize(self):
        for i in self._initializers:
            i.initialize()

        self._is_initialized = True



"""
:mod:`zsl.db.model.raw_model`
-----------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from builtins import object
from zsl.utils.model_helper import update_model


class ModelBase(object):
    def update(self, app_model, forbidden_keys=None, inverse=False):
        """
        Updates the raw model. Consult `zsl.utils.model_helper.update_model`.
        """
        if forbidden_keys is None:
            forbidden_keys = []

        update_model(self, app_model, forbidden_keys, inverse)

    def get_app_model(self, id_name='id', hints=None):
        return self.__app_model__(self.__dict__, id_name, hints)



"""
:mod:`zsl.tasks.asl.schedule_kill_worker_task`
----------------------------------------------
"""
from __future__ import unicode_literals
from builtins import object
from zsl.utils.gearman_helper import schedule_gearman_task


class ScheduleKillWorkerTask(object):
    def perform(self, _):
        # Create gearman.
        schedule_gearman_task('zsl/kill_worker_task', {})
        return 'job submitted'



"""
:mod:`zsl.interface.webservice.web_application_loader`
------------------------------------------------------
"""

from __future__ import unicode_literals
# Now import the application and the remaining stuff.

from zsl.interface.webservice.performers.default import create_not_found_mapping
from zsl.interface.webservice.performers.resource import create_resource_mapping
from zsl.interface.webservice.performers.task import create_web_task
from zsl.interface.webservice.performers.method import call_exposers_in_method_packages


def load():
    _load_performers()


def _load_performers():
    """
    Import in this form is necessary so that we avoid the unwanted behavior and immediate initialization of the
    application objects. This makes the initialization procedure run in the time when it is necessary and has every
    required resources.
    """
    create_not_found_mapping()
    create_resource_mapping()
    create_web_task()
    call_exposers_in_method_packages()



import http.client
from unittest.case import TestCase

from zsl.task.job_context import JobContext, WebJobContext

from zsl.application.containers.web_container import WebContainer
from zsl.interface.webservice.utils.execution import execute_web_task
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration
from zsl import inject, Zsl


class ErrorHandlingTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update(CORS={'origin': 'origin'})
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='ErrorHandlingTestCase', container=WebContainer,
        config_object=CONFIG)

    @inject(app=Zsl)
    def testErrorTaskExecution(self, app):
        with app.test_request_context('/'):
            jc = WebJobContext(None, None, None, None, None)
            JobContext.set_current_context(jc)
            response = execute_web_task(jc, self._throwError)
            self.assertEqual(http.client.INTERNAL_SERVER_ERROR,
                             response.status_code,
                             "Status code must be error.")
            self.assertTrue('Access-Control-Allow-Origin' in response.headers,
                            "CORS must be set up")
            self.assertEqual('origin',
                             response.headers['Access-Control-Allow-Origin'],
                             "Default origin must be returned.")

    def _throwError(self):
        raise Exception("aa")



"""
:mod:`zsl.utils.injection_helper`
---------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

import logging
from typing import Type

from future.utils import viewitems
import injector
from injector import Binder, ClassProvider, BindingKey, reraise, CallError
import inspect
import functools

from injector import Scope

from zsl._state import get_current_app


def instantiate(cls):
    inj = get_current_app().injector

    if hasattr(cls, "__new__"):
        task = inj.create_object(cls)
    else:
        task = cls()

    return task


def inject(**bindings):
    """
    Decorator for injecting parameters for ASL objects.
    """

    def outer_wrapper(f):
        def function_wrapper(ff):
            for key, value in viewitems(bindings):
                bindings[key] = BindingKey(value)

            @functools.wraps(ff)
            def _inject(*args, **kwargs):
                inj = get_current_app().injector
                dependencies = inj.args_to_inject(
                    function=ff,
                    bindings=bindings,
                    owner_key=ff
                )
                dependencies.update(kwargs)
                try:
                    return ff(*args, **dependencies)
                except TypeError as e:
                    reraise(e, CallError(ff, args, dependencies, e))

            return _inject

        '''
        Just a convenience method - delegates everything to wrapper.
        '''

        def method_or_class_wrapper(*a, **kwargs):
            """
            Properly installs the injector into the object so that the injection can be performed.
            """
            inj = get_current_app().injector
            # Install injector into the self instance if this is a method call.
            inj.install_into(a[0])

            # Use the generic wrapper.
            inject_f = injector.inject(**bindings)
            # Call the method.
            return inject_f(f)(*a, **kwargs)

        if inspect.ismethod(f):
            return method_or_class_wrapper
        else:
            return function_wrapper(f)

    return outer_wrapper


def bind(interface, to=None, scope=None):
    logging.getLogger(__name__).warning("Using this style of binding is deprecated. Use Modules instead.")
    get_current_app().injector.binder.bind(interface, to, scope)


def simple_bind(binder, cls, scope):
    # type: (Binder, Type, Scope) -> None
    binder.bind(interface=cls, to=ClassProvider(cls).get(binder.injector), scope=scope)



from unittest.case import TestCase

from flask.config import Config

from zsl.application.containers.container import IoCContainer
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration

from zsl.testing.db import DbTestCase, IN_MEMORY_DB_SETTINGS
from zsl.utils.injection_helper import bind
from zsl.utils.redis_helper import Keymaker
from zsl.utils.testing import set_test_responder
from zsl import inject

__author__ = 'peter'

set_test_responder()


class TestKeymaker(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        'test_redis_helper',
        container=IoCContainer,
        config_object=IN_MEMORY_DB_SETTINGS
    )

    def test_empty(self):
        keymaker = Keymaker({})

        with self.assertRaises(AttributeError):
            keymaker.a()

        self.assertEqual(len(keymaker.__dict__), 0, "There should be no method in keymaker")

    def test_normal(self):
        keymaker = Keymaker({'a': 'AAA', 'b': 'XX'})

        self.assertEqual(keymaker.a(), 'AAA', "Pure method")
        self.assertEqual(keymaker.b('x', 'y'), 'XX:x:y', "Method with arguments")

        self.assertEqual(keymaker.b('x', None, 'y'), 'XX:x:y', "Method with a None argument")

        self.assertEqual(keymaker.b('x', None, 0, False, 'y'),
                         'XX:x:0:False:y',
                         "Method with a None and falsified arguments")

    def test_with_prefix(self):
        keymaker = Keymaker({'a': 'AAA', 'b': 'XX'}, prefix="testing")

        self.assertEqual(keymaker.a(), 'testing:AAA', "Pure method with prefix")
        self.assertEqual(keymaker.b('x', 'y'), 'testing:XX:x:y', "Method with arguments and prefix")

    @inject(config=Config)
    def test_with_global_prefix(self, config):
        # type: (Config)->None
        config.setdefault('REDIS', {'prefix': 'global_prefix'})

        keymaker = Keymaker({'a': 'AAA', 'b': 'XX'}, prefix="testing")

        self.assertEqual(keymaker.a(), 'global_prefix:testing:AAA', "Pure method with global and local prefix")
        self.assertEqual(keymaker.b('x', 'y'),
                         'global_prefix:testing:XX:x:y',
                         "Method with arguments and global and local prefix")
        config['REDIS']['prefix'] = None



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

from zsl.configuration import InvalidConfigurationException


class MethodConfiguration(object):
    def __init__(self, package=None, packages=None, url_prefix='method'):
        if packages is not None and package is not None:
            raise InvalidConfigurationException("Can not take both packages and package in method configuration.")
        packages = tuple(packages) if packages is not None else ()
        self._packages = (package,) if package is not None else packages

        if '/' in url_prefix:
            raise InvalidConfigurationException("MethodConfiguration url_prefix parameter can not contain slashes.")
        self._url_prefix = url_prefix

    @property
    def url_prefix(self):
        return self._url_prefix

    @property
    def packages(self):
        return self._packages



"""
:mod:`zsl.application.modules.logger_module`
--------------------------------------------
"""
from __future__ import unicode_literals

import logging.config

from injector import Binder
from injector import Module, inject

from zsl import Config, Zsl
from zsl import inject


class LoggerModule(Module):
    """Configure the application logger."""

    LOGGING_CONFIG_NAME = 'LOGGING'

    def configure(self, binder):
        # type: (Binder) -> None
        super(LoggerModule, self).configure(binder)
        self.configure_logging()

    @inject(config=Config, app=Zsl)
    def configure_logging(self, config, app):
        # type: (Config) -> None
        default_config = dict(
            version=1,
            root=dict(
                level='DEBUG' if config.get('DEBUG', False) else 'WARNING'
            )
        )
        logging.config.dictConfig(config.get(LoggerModule.LOGGING_CONFIG_NAME, default_config))
        self._recreate_app_logger(app)

    def _recreate_app_logger(self, app):
        logging._acquireLock()
        del logging.getLogger(app.logger_name).manager.loggerDict[app.logger_name]
        logging._releaseLock()
        app._logger = logging.getLogger(app.logger_name)



from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import *

from typing import Callable
from unittest.case import TestCase

from zsl.application.containers.web_container import WebContainer
from zsl.tasks.zsl.version_task import VersionTask
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration

from zsl.errors import ZslError
from zsl.interface.task import create_task


class ExecTaskTestCase(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='ExecTaskTestCase', container=WebContainer,
        config_object=IN_MEMORY_DB_SETTINGS)

    def testCreateTaskNone(self):
        with self.assertRaises(ZslError):
            create_task(None)

    def testCreateTaskValidCls(self):
        task, task_callable = create_task(VersionTask)
        self.assertIsInstance(task, VersionTask,
                              "A proper task must be returned.")
        self.assertIsInstance(task_callable, Callable,
                              "A proper callable must be returned.")

    def testCreateTaskValidPath(self):
        task, task_callable = create_task("zsl/version_task")
        self.assertIsInstance(task, VersionTask,
                              "A proper task must be returned.")
        self.assertIsInstance(task_callable, Callable,
                              "A proper callable must be returned.")



"""
:mod:`zsl.utils.xml_helper` -- xml helpers
==========================================

Helper functions for working with XML and ElementTree.
"""
from __future__ import unicode_literals
from future.utils import viewitems

import xml.etree.cElementTree as ET
import requests
from functools import reduce


class NotValidXmlException(Exception):
    """Exception raised on invalid xml"""
    # TODO try to use some build in exception #13299
    pass


def required_attributes(element, *attributes):
    """Check element for required attributes. Raise ``NotValidXmlException`` on error.

    :param element: ElementTree element
    :param attributes: list of attributes names to check
    :raises NotValidXmlException: if some argument is missing
    """
    if not reduce(lambda still_valid, param: still_valid and param in element.attrib, attributes, True):
        raise NotValidXmlException(msg_err_missing_attributes(element.tag, *attributes))


def required_elements(element, *children):
    """Check element (``xml.etree.ElementTree.Element``) for required children, defined as XPath. Raise
    ``NotValidXmlException`` on error.

    :param element: ElementTree element
    :param children: list of XPaths to check
    :raises NotValidXmlException: if some child is missing
    """
    for child in children:
        if element.find(child) is None:
            raise NotValidXmlException(msg_err_missing_children(element.tag, *children))


def required_items(element, children, attributes):
    """Check an xml element to include given attributes and children.

    :param element: ElementTree element
    :param children: list of XPaths to check
    :param attributes: list of attributes names to check
    :raises NotValidXmlException: if some argument is missing
    :raises NotValidXmlException: if some child is missing
    """
    required_elements(element, *children)
    required_attributes(element, *attributes)


def msg_err_missing_attributes(tag, *attributes):
    """Format message for missing attributes.

    :param tag: tag name
    :param attributes: list of attributes
    :return: message
    :rtype: str
    """
    return "Missing one or more required attributes (%s) in xml tag %s" % ('|'.join(attributes), tag)


def msg_err_missing_children(tag, *children):
    """Format message for missing children.

    :param tag: tag name
    :param children: list of children
    :return: message
    :rtype: str
    """
    return "Missing one or more required children (%s) in xml tag %s" % ('|'.join(children), tag)


def attrib_to_dict(element, *args, **kwargs):
    """For an ElementTree ``element`` extract specified attributes. If an attribute does not exists, its value will be
    ``None``.

    attrib_to_dict(element, 'attr_a', 'attr_b') -> {'attr_a': 'value', 'attr_a': 'value'}

    Mapping between xml attributes and dictionary keys is done with kwargs.
    attrib_to_dict(element, my_new_name = 'xml_atribute_name', ..)
    """
    if len(args) > 0:
        return {key: element.get(key) for key in args}

    if len(kwargs) > 0:
        return {new_key: element.get(old_key) for new_key, old_key in viewitems(kwargs)}

    return element.attrib


def get_xml_root(xml_path):
    """Load and parse an xml by given xml_path and return its root.

    :param xml_path: URL to a xml file
    :type xml_path: str
    :return: xml root
    """
    tree = ET.fromstring(requests.get(xml_path).content)
    return tree.getroot()


def element_to_int(element, attribute=None):
    """Convert ``element`` object to int. If attribute is not given, convert ``element.text``.

    :param element: ElementTree element
    :param attribute: attribute name
    :type attribute: str
    :returns: integer
    :rtype: int
    """
    if attribute is not None:
        return int(element.get(attribute))
    else:
        return int(element.text)


def create_el(name, text=None, attrib=None):
    """Create element with given attributes and set element.text property to given
    text value (if text is not None)

    :param name: element name
    :type name: str
    :param text: text node value
    :type text: str
    :param attrib: attributes
    :type attrib: dict
    :returns: xml element
    :rtype: Element
    """
    if attrib is None:
        attrib = {}

    el = ET.Element(name, attrib)
    if text is not None:
        el.text = text
    return el
